from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import DomainModel, DomainModelApply, TypeApplyList, TypeList

if TYPE_CHECKING:
    from ._date_transformation import DateTransformationApply

__all__ = ["ShopTransformation", "ShopTransformationApply", "ShopTransformationList", "ShopTransformationApplyList"]


class ShopTransformation(DomainModel):
    space: ClassVar[str] = "power-ops"
    type_name: Optional[str] = Field(None, alias="typeName")
    end: Optional[list[str]] = None
    start: Optional[list[str]] = None

    def as_apply(self) -> ShopTransformationApply:
        return ShopTransformationApply(
            external_id=self.external_id,
            type_name=self.type_name,
            end=self.end,
            start=self.start,
        )


class ShopTransformationApply(DomainModelApply):
    space: ClassVar[str] = "power-ops"
    type_name: Optional[str] = None
    end: Union[list[DateTransformationApply], list[str], None] = Field(default=None, repr=False)
    start: Union[list[DateTransformationApply], list[str], None] = Field(default=None, repr=False)

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        properties = {}
        if self.type_name is not None:
            properties["typeName"] = self.type_name
        if properties:
            source = dm.NodeOrEdgeData(
                source=dm.ContainerId("power-ops", "ShopTransformation"),
                properties=properties,
            )
            sources.append(source)
        if sources:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.existing_version,
                sources=sources,
            )
            nodes = [this_node]
        else:
            nodes = []

        edges = []
        cache.add(self.external_id)

        for end in self.end or []:
            edge = self._create_end_edge(end)
            if edge.external_id not in cache:
                edges.append(edge)
                cache.add(edge.external_id)

            if isinstance(end, DomainModelApply):
                instances = end._to_instances_apply(cache)
                nodes.extend(instances.nodes)
                edges.extend(instances.edges)

        for start in self.start or []:
            edge = self._create_start_edge(start)
            if edge.external_id not in cache:
                edges.append(edge)
                cache.add(edge.external_id)

            if isinstance(start, DomainModelApply):
                instances = start._to_instances_apply(cache)
                nodes.extend(instances.nodes)
                edges.extend(instances.edges)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))

    def _create_end_edge(self, end: Union[str, DateTransformationApply]) -> dm.EdgeApply:
        if isinstance(end, str):
            end_node_ext_id = end
        elif isinstance(end, DomainModelApply):
            end_node_ext_id = end.external_id
        else:
            raise TypeError(f"Expected str or DateTransformationApply, got {type(end)}")

        return dm.EdgeApply(
            space="power-ops",
            external_id=f"{self.external_id}:{end_node_ext_id}",
            type=dm.DirectRelationReference("power-ops", "ShopTransformation.end"),
            start_node=dm.DirectRelationReference(self.space, self.external_id),
            end_node=dm.DirectRelationReference("power-ops", end_node_ext_id),
        )

    def _create_start_edge(self, start: Union[str, DateTransformationApply]) -> dm.EdgeApply:
        if isinstance(start, str):
            end_node_ext_id = start
        elif isinstance(start, DomainModelApply):
            end_node_ext_id = start.external_id
        else:
            raise TypeError(f"Expected str or DateTransformationApply, got {type(start)}")

        return dm.EdgeApply(
            space="power-ops",
            external_id=f"{self.external_id}:{end_node_ext_id}",
            type=dm.DirectRelationReference("power-ops", "ShopTransformation.start"),
            start_node=dm.DirectRelationReference(self.space, self.external_id),
            end_node=dm.DirectRelationReference("power-ops", end_node_ext_id),
        )


class ShopTransformationList(TypeList[ShopTransformation]):
    _NODE = ShopTransformation

    def as_apply(self) -> ShopTransformationApplyList:
        return ShopTransformationApplyList([node.as_apply() for node in self.data])


class ShopTransformationApplyList(TypeApplyList[ShopTransformationApply]):
    _NODE = ShopTransformationApply



from __future__ import annotations

from typing import ClassVar, Optional

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import DomainModel, DomainModelApply, TypeApplyList, TypeList

__all__ = ["Point", "PointApply", "PointList", "PointApplyList"]


class Point(DomainModel):
    space: ClassVar[str] = "power-ops"
    position: Optional[int] = None
    quantity: Optional[float] = None
    minimum_quantity: Optional[float] = Field(None, alias="minimumQuantity")
    price_amount: Optional[float] = Field(None, alias="priceAmount")
    energy_price_amount: Optional[float] = Field(None, alias="energyPriceAmount")

    def as_apply(self) -> PointApply:
        return PointApply(
            external_id=self.external_id,
            position=self.position,
            quantity=self.quantity,
            minimum_quantity=self.minimum_quantity,
            price_amount=self.price_amount,
            energy_price_amount=self.energy_price_amount,
        )


class PointApply(DomainModelApply):
    space: ClassVar[str] = "power-ops"
    position: Optional[int] = None
    quantity: Optional[float] = None
    minimum_quantity: Optional[float] = None
    price_amount: Optional[float] = None
    energy_price_amount: Optional[float] = None

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        properties = {}
        if self.position is not None:
            properties["position"] = self.position
        if self.quantity is not None:
            properties["quantity"] = self.quantity
        if self.minimum_quantity is not None:
            properties["minimumQuantity"] = self.minimum_quantity
        if self.price_amount is not None:
            properties["priceAmount"] = self.price_amount
        if self.energy_price_amount is not None:
            properties["energyPriceAmount"] = self.energy_price_amount
        if properties:
            source = dm.NodeOrEdgeData(
                source=dm.ContainerId("power-ops", "Point"),
                properties=properties,
            )
            sources.append(source)
        if sources:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.existing_version,
                sources=sources,
            )
            nodes = [this_node]
        else:
            nodes = []

        edges = []
        cache.add(self.external_id)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))


class PointList(TypeList[Point]):
    _NODE = Point

    def as_apply(self) -> PointApplyList:
        return PointApplyList([node.as_apply() for node in self.data])


class PointApplyList(TypeApplyList[PointApply]):
    _NODE = PointApply



from __future__ import annotations

import datetime
from typing import ClassVar, Optional

from cognite.client import data_modeling as dm

from ._core import DomainModel, DomainModelApply, TypeApplyList, TypeList

__all__ = ["DateTimeInterval", "DateTimeIntervalApply", "DateTimeIntervalList", "DateTimeIntervalApplyList"]


class DateTimeInterval(DomainModel):
    space: ClassVar[str] = "power-ops"
    start: Optional[datetime.datetime] = None
    end: Optional[datetime.datetime] = None

    def as_apply(self) -> DateTimeIntervalApply:
        return DateTimeIntervalApply(
            external_id=self.external_id,
            start=self.start,
            end=self.end,
        )


class DateTimeIntervalApply(DomainModelApply):
    space: ClassVar[str] = "power-ops"
    start: Optional[datetime.datetime] = None
    end: Optional[datetime.datetime] = None

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        properties = {}
        if self.start is not None:
            properties["start"] = self.start.isoformat()
        if self.end is not None:
            properties["end"] = self.end.isoformat()
        if properties:
            source = dm.NodeOrEdgeData(
                source=dm.ContainerId("power-ops", "DateTimeInterval"),
                properties=properties,
            )
            sources.append(source)
        if sources:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.existing_version,
                sources=sources,
            )
            nodes = [this_node]
        else:
            nodes = []

        edges = []
        cache.add(self.external_id)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))


class DateTimeIntervalList(TypeList[DateTimeInterval]):
    _NODE = DateTimeInterval

    def as_apply(self) -> DateTimeIntervalApplyList:
        return DateTimeIntervalApplyList([node.as_apply() for node in self.data])


class DateTimeIntervalApplyList(TypeApplyList[DateTimeIntervalApply]):
    _NODE = DateTimeIntervalApply



from __future__ import annotations

import datetime
import secrets
from collections.abc import Sequence
from typing import Literal, Optional, overload
from urllib.parse import urlparse
from uuid import uuid4

import arrow
import requests
from cognite.client import CogniteClient
from cognite.client.data_classes import UserProfile, filters
from cognite.client.data_classes.events import EventSort
from cognite.client.exceptions import CogniteAPIError

from cognite.powerops.client.shop.shop_run_filter import SHOPRunFilter

from .shop_run import SHOPRun, ShopRunEvent, SHOPRunList

DEFAULT_READ_LIMIT = 25


def _unique_short_str(nbytes: int) -> str:
    return secrets.token_hex(nbytes=nbytes)


class SHOPRunAPI:
    def __init__(self, client: CogniteClient, dataset_id: int, cogshop_version: str):
        self._cdf = client
        self._dataset_id = dataset_id
        self.cogshop_version = cogshop_version

    def trigger_case(self, case_file: str, watercourse: str, source: str | None = None) -> SHOPRun:
        """
        Trigger SHOP for a given case file.

        Args:
            case_file: Case file as a string. Expected to be YAML.
            watercourse: The watercourse to run SHOP this case file is for.
            source: The source of the SHOP trigger. If nothing is passed, the method will
                try to detect the service principal of the current user.

        Returns:
            The new SHOP run created.
        """
        user: UserProfile | None
        try:
            user = self._cdf.iam.user_profiles.me()
        except CogniteAPIError as e:
            if e.code == 404:
                user = None
            else:
                raise

        if source is None and user:
            source = user.user_identifier
        elif source is None:
            source = "manual"

        if user:
            display_name = user.display_name
        else:
            display_name = "Unknown"

        meta = self._cdf.files.upload_bytes(
            # On Windows machines, some bytes can get lost in the encoding process
            # when uploading a string to CDF. This is a workaround.
            content=case_file.encode("utf-8"),
            name=f"Manual Case by {display_name}",
            mime_type="application/yaml",
            external_id=f"cog_shop_manual/{uuid4()!s}",
            metadata={"shop:type": "cog_shop_case", "user": display_name},
            data_set_id=self._dataset_id,
            source=source,
        )
        now = datetime.datetime.now(datetime.timezone.utc)
        now_isoformat = now.isoformat().replace("+00:00", "Z")

        new_event = SHOPRun(
            external_id=f"SHOP_RUN_{now_isoformat}_{_unique_short_str(3)}",
            watercourse=watercourse,
            start=now,
            end=None,
            _case_file_external_id=meta.external_id,
            _shop_files=[],
            shop_version=self.cogshop_version,
            _client=self._cdf,
            source=source,
        )
        self._cdf.events.create(new_event.as_cdf_event(self._dataset_id))

        self._trigger_shop_container(new_event)
        return new_event

    def _trigger_shop_container(self, event: SHOPRun):
        def auth(r: requests.PreparedRequest) -> requests.PreparedRequest:
            auth_header_name, auth_header_value = self._cdf._config.credentials.authorization_header()
            r.headers[auth_header_name] = auth_header_value
            return r

        response = requests.post(
            url=self._shop_url(),
            json={"shopEventExternalId": event.external_id, "cogShopVersion": self.cogshop_version},
            auth=auth,
        )
        response.raise_for_status()

    def _shop_url(self) -> str:
        project = self._cdf.config.project

        cluster = urlparse(self._cdf.config.base_url).netloc.split(".", 1)[0]

        environment = project.split("-")[-1]
        if environment in {"dev", "staging"}:
            stage = ".staging"
        elif environment == "prod":
            stage = ""
        else:
            raise ValueError(f"Can't detect prod/staging from project name: {project!r}")

        return f"https://power-ops-api{stage}.{cluster}.cognite.ai/{project}/run-shop"

    def _load_cdf_event_shop_runs(
        self,
        extra_filters: Optional[list[filters.Filter]] = None,
        limit: int = DEFAULT_READ_LIMIT,
        event_sort: EventSort = None,
    ) -> SHOPRunList:
        is_type = filters.Equals("type", ShopRunEvent.event_type)
        events = self._cdf.events.filter(filters.And(is_type, *extra_filters), limit=limit, sort=event_sort)
        return SHOPRunList.load(events)

    @overload
    def retrieve(self, external_id: str) -> SHOPRun | None:
        ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> SHOPRunList:
        ...

    def retrieve(
        self, external_id: str | Sequence[str], ignore_unknown_ids: bool = True
    ) -> SHOPRun | SHOPRunList | None:
        """
        Retrieve a SHOP run.

        Args:
            external_id: The external id(s) of the SHOP run(s) to retrieve.
            ignore_unknown_ids: Whether to ignore unknown ids or raise an error.

        Returns:
            The SHOP run(s). None if it is a single external id, and it is not found.
        """
        if isinstance(external_id, str):
            event = self._cdf.events.retrieve(external_id=external_id)
            if event is None:
                return None
            return SHOPRun.load(event)
        elif isinstance(external_id, Sequence):
            return SHOPRunList.load(
                self._cdf.events.retrieve_multiple(external_ids=external_id, ignore_unknown_ids=ignore_unknown_ids)
            )
        else:
            raise TypeError(f"Invalid type {type(external_id)} for external_id.")

    def retrieve_latest(
        self,
        watercourse: str | list[str] | None = None,
        source: str | list[str] | None = None,
        start_after: str | arrow.Arrow | datetime.datetime | None = None,
        start_before: str | arrow.Arrow | datetime.datetime | None = None,
        end_after: str | arrow.Arrow | datetime.datetime | None = None,
        end_before: str | arrow.Arrow | datetime.datetime | None = None,
        latest_by: Literal[
            "start_time",
            "end_time",
            "created_time",
            "last_updated_time",
        ] = "last_updated_time",
    ) -> SHOPRun | None:
        """
        Retrieves the the latest shop run given the filters.

         Args:
            watercourse: The watercourse(s) to filter on.
            source: The source(s) to filter on.
            start_after: The start time after which the SHOP run must have started.
            start_before: The start time before which the SHOP run must have started.
            end_after: The end time after which the SHOP run must have ended.
            end_before: The end time before which the SHOP run must have ended.
            latest_by: The property to retrieve by . The most reliable is `last_updated_time` and `created_time`.
        Returns:
            The most recent SHOP run, None if nothing matched the filter
        """
        extra_filters = SHOPRunFilter(
            watercourse=watercourse,
            source=source,
            start_after=start_after,
            start_before=start_before,
            end_after=end_after,
            end_before=end_before,
        ).get_filters()
        event_sort = EventSort(property=latest_by, order="desc")

        shop_runs = self._load_cdf_event_shop_runs(extra_filters=extra_filters, limit=1, event_sort=event_sort)
        if len(shop_runs) == 0:
            return None

        return shop_runs[0]

    def list(
        self,
        watercourse: str | list[str] | None = None,
        source: str | list[str] | None = None,
        start_after: str | arrow.Arrow | datetime.datetime | None = None,
        start_before: str | arrow.Arrow | datetime.datetime | None = None,
        end_after: str | arrow.Arrow | datetime.datetime | None = None,
        end_before: str | arrow.Arrow | datetime.datetime | None = None,
        limit: int = DEFAULT_READ_LIMIT,
    ) -> SHOPRunList:
        """List the filtered SHOP runs.
        Provide time stamps strings as YYYY-MM-DD) or otherwise something parsable by `arrow.get()` can parse.

        Args:
            watercourse: The watercourse(s) to filter on.
            source: The source(s) to filter on. WARNING: Most
            start_after: The start time after which the SHOP run must have started.
            start_before: The start time before which the SHOP run must have started.
            end_after: The end time after which the SHOP run must have ended.
            end_before: The end time before which the SHOP run must have ended.
            limit: The maximum number of SHOP runs to return.

        Returns:
            A list of SHOP runs.
        """

        extra_filters = SHOPRunFilter(
            watercourse=watercourse,
            source=source,
            start_after=start_after,
            start_before=start_before,
            end_after=end_after,
            end_before=end_before,
        ).get_filters()
        return self._load_cdf_event_shop_runs(extra_filters=extra_filters, limit=limit)



from __future__ import annotations

from collections.abc import Sequence
from typing import overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm

from cognite.powerops.client._generated.cogshop1.data_classes import (
    CommandsConfig,
    CommandsConfigApply,
    CommandsConfigApplyList,
    CommandsConfigList,
)

from ._core import DEFAULT_LIMIT_READ, TypeAPI


class CommandsConfigAPI(TypeAPI[CommandsConfig, CommandsConfigApply, CommandsConfigList]):
    def __init__(self, client: CogniteClient, view_id: dm.ViewId):
        super().__init__(
            client=client,
            sources=view_id,
            class_type=CommandsConfig,
            class_apply_type=CommandsConfigApply,
            class_list=CommandsConfigList,
        )
        self.view_id = view_id

    def apply(
        self, commands_config: CommandsConfigApply | Sequence[CommandsConfigApply], replace: bool = False
    ) -> dm.InstancesApplyResult:
        if isinstance(commands_config, CommandsConfigApply):
            instances = commands_config.to_instances_apply()
        else:
            instances = CommandsConfigApplyList(commands_config).to_instances_apply()
        return self._client.data_modeling.instances.apply(nodes=instances.nodes, edges=instances.edges, replace=replace)

    def delete(self, external_id: str | Sequence[str]) -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(CommandsConfigApply.space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(CommandsConfigApply.space, id) for id in external_id],
            )

    @overload
    def retrieve(self, external_id: str) -> CommandsConfig:
        ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> CommandsConfigList:
        ...

    def retrieve(self, external_id: str | Sequence[str]) -> CommandsConfig | CommandsConfigList:
        if isinstance(external_id, str):
            return self._retrieve((self.sources.space, external_id))
        else:
            return self._retrieve([(self.sources.space, ext_id) for ext_id in external_id])

    def list(
        self,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> CommandsConfigList:
        filter_ = _create_filter(
            self.view_id,
            external_id_prefix,
            filter,
        )

        return self._list(limit=limit, filter=filter_)


def _create_filter(
    view_id: dm.ViewId,
    external_id_prefix: str | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if external_id_prefix:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



import inspect
import sys
from abc import ABC, abstractmethod
from datetime import datetime, timedelta
from logging import getLogger
from typing import Any, Literal

import numpy as np
import pandas as pd
from cognite.client import CogniteClient
from cognite.client.utils import ms_to_datetime
from pydantic import BaseModel
from typing_extensions import Self

from cognite.powerops.utils.cdf.calls import retrieve_range

logger = getLogger(__name__)


class RelativeDatapoint(BaseModel):
    offset_minute: float
    offset_value: float


def _relative_datapoints_to_series(
    relative_datapoints: list[RelativeDatapoint], start_time: datetime, shift_minutes: int = 0
) -> pd.Series:
    """Converts 'relative datapoints' to a pandas Series

    Example:
        >>> start_time = datetime(2000, 1, 1, 12)
        >>> relative_datapoints = [
        ...     RelativeDatapoint(offset_minute=0, offset_value=42),
        ...     RelativeDatapoint(offset_minute=60, offset_value=420),
        ...     RelativeDatapoint(offset_minute=1440, offset_value=4200),
        ... ]
        >>> _relative_datapoints_to_series(relative_datapoints, start_time)
        2000-01-01 12:00:00      42.0
        2000-01-01 13:00:00     420.0
        2000-01-02 12:00:00    4200.0
        dtype: float64
    """
    datapoint_values = [dp.offset_value for dp in relative_datapoints]  # consider creating class method for this
    return pd.Series(
        datapoint_values,
        index=[start_time + timedelta(minutes=int(dp.offset_minute) + shift_minutes) for dp in relative_datapoints],
    )


class Transformation(BaseModel, ABC):
    @classmethod
    def load(cls, transformation_: dict[str, Any]) -> Self:
        (transformation_name,) = transformation_
        if transformation_body := transformation_.get(transformation_name):
            return _TRANSFORMATIONS_BY_CLASS_NAME[transformation_name].__call__(**transformation_body["input"])
        else:
            return _TRANSFORMATIONS_BY_CLASS_NAME[transformation_name].__call__()

    @abstractmethod
    def apply(
        self,
        time_series_data: pd.Series,
    ):
        ...


class DynamicTransformation(Transformation):
    @abstractmethod
    def pre_apply(self, client: CogniteClient, shop_model: dict, start: int, end: int):
        ...


class Add(Transformation):
    value: float

    def apply(
        self,
        time_series_data: pd.Series,
    ):
        return time_series_data + self.value


class Multiply(Transformation):
    value: float

    def apply(
        self,
        time_series_data: pd.Series,
    ):
        return time_series_data * self.value


class StaticValues(DynamicTransformation):
    shift_minutes: int = 0  # This could be given at runtime - looks like default value of 0 is always used for now
    relative_datapoints: list[RelativeDatapoint]
    pre_apply_has_run: bool = False
    _start: int

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, value: int):
        self._start = value

    def pre_apply(self, client: CogniteClient, shop_model: dict, start: int, end: int):
        self.start = start
        self.pre_apply_has_run = True

    def apply(self, _: pd.Series) -> pd.Series:
        if not self.pre_apply_has_run:
            raise ValueError("pre_apply function has not run - missing neccessary properties to run transformation")
        return _relative_datapoints_to_series(self.relative_datapoints, ms_to_datetime(self.start), self.shift_minutes)


class ToBool(Transformation):
    def apply(self, time_series_data: pd.Series) -> pd.Series:
        return (time_series_data > 0).astype(int)


class ZeroIfNotOne(Transformation):
    def apply(self, time_series_data: pd.Series) -> pd.Series:
        return (time_series_data == 1).astype(int)


class OneIfTwo(Transformation):
    def apply(self, time_series_data: pd.Series) -> pd.Series:
        return (time_series_data == 2).astype(int)


class HeightToVolume(DynamicTransformation):
    object_type: str
    object_name: str
    pre_apply_has_run: bool = False
    _volumes: list[float]
    _heights: list[float]

    @property
    def volumes(self):
        return self._volumes

    @volumes.setter
    def volumes(self, value: list[float]):
        self._volumes = value

    @property
    def heights(self):
        return self._volumes

    @heights.setter
    def heights(self, value: list[float]):
        self._heights = value

    @staticmethod
    def height_to_volume(time_series_data: pd.Series, heights: list[float], volumes: list[float]) -> pd.Series:
        def interpolate(height: float) -> float:
            """Height to volume"""
            if height < min(heights) or max(heights) < height:
                logger.warning(f"Outside interpoaltion bounds [{min(heights)}, {max(heights)}]. Got {height}.")
            return float(np.interp(height, heights, volumes))

        return time_series_data.map(interpolate)

    def pre_apply(self, client: CogniteClient, shop_model: dict, start: int, end: int):
        self.volumes = shop_model[self.object_type][self.object_name]["vol_head"]["x"]
        self.heights = shop_model[self.object_type][self.object_name]["vol_head"]["y"]
        self.pre_apply_has_run = True

    def apply(self, time_series_data: pd.Series) -> pd.Series:
        if self.pre_apply_has_run:
            return self.height_to_volume(time_series_data, self.heights, self.volumes)
        else:
            raise ValueError("pre_apply function has not run - missing neccessary properties to run transformation")


class DoNothing(Transformation):
    def apply(self, time_series_data: pd.Series) -> pd.Series:
        return time_series_data


class AddFromOffset(Transformation):
    shift_minutes: int = 0
    relative_datapoints: list[RelativeDatapoint]

    def apply(self, time_series_data: pd.Series) -> pd.Series:
        first_timestamp = min(time_series_data.index)
        non_relative_datapoints = _relative_datapoints_to_series(
            self.relative_datapoints, first_timestamp, self.shift_minutes
        )
        union_index = time_series_data.index.union(non_relative_datapoints.index)
        # fillna(0) since we are adding
        non_relative_datapoints = non_relative_datapoints.reindex(union_index).ffill().fillna(0)
        # If a timestamp does not exist in the original datapoints we add that timestamp with a forwardfilled value
        return time_series_data.reindex(union_index).ffill() + non_relative_datapoints


class MultiplyFromOffset(Transformation):
    shift_minutes: int = 0
    relative_datapoints: list[RelativeDatapoint]

    def apply(self, time_series_data: pd.Series) -> pd.Series:
        first_timestamp = min(time_series_data.index)
        non_relative_datapoints = _relative_datapoints_to_series(
            self.relative_datapoints, first_timestamp, self.shift_minutes
        )
        union_index = time_series_data.index.union(non_relative_datapoints.index)
        # fillna(1) since we are multiplying
        non_relative_datapoints = non_relative_datapoints.reindex(union_index).ffill().fillna(1)
        # If a timestamp does not exist in the original datapoints we add that timestamp with a forwardfilled value
        return time_series_data.reindex(union_index).ffill() * non_relative_datapoints


class AddWaterInTransit(DynamicTransformation, arbitrary_types_allowed=True):
    discharge_ts_external_id: str
    transit_object_type: Literal["plant", "gate"]
    transit_object_name: str
    pre_apply_has_run: bool = False
    _start: int
    _end: int
    _shape: dict[int, float]
    _discharge: pd.Series

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, value: int):
        self._start = value

    @property
    def end(self):
        return self._end

    @end.setter
    def end(self, value: int):
        self._end = value

    @property
    def shape(self):
        return self._shape

    @shape.setter
    def shape(self, value: dict[int, float]):
        self._shape = value

    @property
    def discharge(self):
        return self._discharge

    @discharge.setter
    def discharge(self, value: pd.Series):
        self._discharge = value

    @staticmethod
    def get_shape(model: dict, transit_object_type: str, transit_object_name: str) -> dict[int, float]:
        """object_type must be plant or gate"""
        gate_or_plant = model[transit_object_type][transit_object_name]  # Get description of gate/plant

        # Get shape and time_delay values as Dict[delay,water_percentage]
        if "shape_discharge" in gate_or_plant:
            shape = dict(zip(gate_or_plant["shape_discharge"]["x"], gate_or_plant["shape_discharge"]["y"]))
        elif "time_delay" in gate_or_plant:
            shape = {gate_or_plant["time_delay"]: 1}
        else:
            raise ValueError(
                f"{transit_object_type}.{transit_object_name} does not have a shape_discharge or time_delay attribute"
            )

        return shape

    def pre_apply(self, client: CogniteClient, shop_model: dict, start: int, end: int):
        self.start = start
        self.end = end

        self.shape = self.get_shape(
            model=shop_model, object_type=self.transit_object_type, object_name=self.transit_object_name
        )

        longest_delay = max(self.shape)  # longest delay in minutes
        longest_delay_ms = 60 * 1000 * longest_delay  # longest delay in milliseconds

        discharge = retrieve_range(  # Get discharge datapoints from time-series
            client=client,
            external_ids=[self.extra_ts_external_id],
            start=self.start - longest_delay_ms,  # Shift start time based on longest delay
            end=self.start,
        )[self.discharge_ts_external_id]

        if discharge.empty:
            logger.warning("Cannot add 'water in transit' - did not get any 'discharge' datapoints!")

        self.discharge = discharge
        self.pre_apply_has_run = True

    @staticmethod
    def add_water_in_transit(
        inflow: pd.Series,
        discharge: pd.Series,
        shape: dict[int, float],
        start: datetime,
        end: datetime,
    ) -> pd.Series:
        """Adds water in transit (previously discharged water) to the inflow time series.

        Args:
            shape (Dict[int, float]): Description of "how much water is delayed by what amount of time".
                e.g.
                shape = {
                    0: 0,       # 0% instantly
                    120: 0.5,   # 50% delayed for 2h
                    1320: 0.5,  # 50% delayed for 22h
                }

        """
        # Forward fill discharge for all (hour) timestamps until start
        one_hour = pd.Timedelta("1h")
        if start - one_hour not in discharge.index:
            discharge[start - one_hour] = np.NaN
        discharge = discharge.resample(one_hour).ffill().ffill()

        # Make sure inflow has values for all (hour) timestamps up until end (exclusive)
        if end not in inflow.index:
            inflow[end] = np.NaN
        inflow = inflow.resample(one_hour).ffill().ffill()

        for delay, water_percentage in shape.items():
            delayed_discharge = discharge.shift(delay, freq="min") * water_percentage
            inflow = inflow.add(delayed_discharge, fill_value=0)

        # Only include delayed discharge that will affect the selected time range
        between_start_and_end = (start <= inflow.index) & (inflow.index < end)
        return inflow.loc[between_start_and_end]

    def apply(
        self,
        time_series_data: pd.Series,
    ) -> pd.Series:
        if time_series_data.empty:
            return time_series_data

        if self.pre_apply_has_run:
            return self.add_water_in_transit(
                inflow=time_series_data,
                discharge=self.discharge,
                shape=self.shape,
                start=ms_to_datetime(self.start),
                end=ms_to_datetime(self.end),
            )
        else:
            raise ValueError("pre_apply function has not run - missing neccessary properties to run transformation")


_TRANSFORMATIONS_BY_CLASS_NAME = dict(
    inspect.getmembers(sys.modules[__name__], lambda member: inspect.isclass(member) and member.__module__ == __name__)
)



from __future__ import annotations

from collections.abc import Sequence
from typing import overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm

from cognite.powerops.client._generated.data_classes import (
    ScenarioTemplate,
    ScenarioTemplateApply,
    ScenarioTemplateApplyList,
    ScenarioTemplateList,
)

from ._core import DEFAULT_LIMIT_READ, TypeAPI


class ScenarioTemplateAPI(TypeAPI[ScenarioTemplate, ScenarioTemplateApply, ScenarioTemplateList]):
    def __init__(self, client: CogniteClient, view_id: dm.ViewId):
        super().__init__(
            client=client,
            sources=view_id,
            class_type=ScenarioTemplate,
            class_apply_type=ScenarioTemplateApply,
            class_list=ScenarioTemplateList,
        )
        self.view_id = view_id

    def apply(
        self, scenario_template: ScenarioTemplateApply | Sequence[ScenarioTemplateApply], replace: bool = False
    ) -> dm.InstancesApplyResult:
        if isinstance(scenario_template, ScenarioTemplateApply):
            instances = scenario_template.to_instances_apply()
        else:
            instances = ScenarioTemplateApplyList(scenario_template).to_instances_apply()
        return self._client.data_modeling.instances.apply(nodes=instances.nodes, edges=instances.edges, replace=replace)

    def delete(self, external_id: str | Sequence[str]) -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(ScenarioTemplateApply.space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(ScenarioTemplateApply.space, id) for id in external_id],
            )

    @overload
    def retrieve(self, external_id: str) -> ScenarioTemplate:
        ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> ScenarioTemplateList:
        ...

    def retrieve(self, external_id: str | Sequence[str]) -> ScenarioTemplate | ScenarioTemplateList:
        if isinstance(external_id, str):
            return self._retrieve((self.sources.space, external_id))
        else:
            return self._retrieve([(self.sources.space, ext_id) for ext_id in external_id])

    def list(
        self,
        watercourse: str | list[str] | None = None,
        watercourse_prefix: str | None = None,
        shop_version: str | list[str] | None = None,
        shop_version_prefix: str | None = None,
        template_version: str | list[str] | None = None,
        template_version_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> ScenarioTemplateList:
        filter_ = _create_filter(
            self.view_id,
            watercourse,
            watercourse_prefix,
            shop_version,
            shop_version_prefix,
            template_version,
            template_version_prefix,
            external_id_prefix,
            filter,
        )

        return self._list(limit=limit, filter=filter_)


def _create_filter(
    view_id: dm.ViewId,
    watercourse: str | list[str] | None = None,
    watercourse_prefix: str | None = None,
    shop_version: str | list[str] | None = None,
    shop_version_prefix: str | None = None,
    template_version: str | list[str] | None = None,
    template_version_prefix: str | None = None,
    external_id_prefix: str | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if watercourse and isinstance(watercourse, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("watercourse"), value=watercourse))
    if watercourse and isinstance(watercourse, list):
        filters.append(dm.filters.In(view_id.as_property_ref("watercourse"), values=watercourse))
    if watercourse_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("watercourse"), value=watercourse_prefix))
    if shop_version and isinstance(shop_version, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("shopVersion"), value=shop_version))
    if shop_version and isinstance(shop_version, list):
        filters.append(dm.filters.In(view_id.as_property_ref("shopVersion"), values=shop_version))
    if shop_version_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("shopVersion"), value=shop_version_prefix))
    if template_version and isinstance(template_version, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("templateVersion"), value=template_version))
    if template_version and isinstance(template_version, list):
        filters.append(dm.filters.In(view_id.as_property_ref("templateVersion"), values=template_version))
    if template_version_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("templateVersion"), value=template_version_prefix))
    if external_id_prefix:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from ._benchmark_bid import BenchmarkBid, BenchmarkBidApply, BenchmarkBidApplyList, BenchmarkBidList
from ._benchmark_process import BenchmarkProcess, BenchmarkProcessApply, BenchmarkProcessApplyList, BenchmarkProcessList
from ._bid_matrix_generator import (
    BidMatrixGenerator,
    BidMatrixGeneratorApply,
    BidMatrixGeneratorApplyList,
    BidMatrixGeneratorList,
)
from ._bid_time_series import BidTimeSeries, BidTimeSeriesApply, BidTimeSeriesApplyList, BidTimeSeriesList
from ._command_config import CommandConfig, CommandConfigApply, CommandConfigApplyList, CommandConfigList
from ._date_time_interval import (
    DateTimeInterval,
    DateTimeIntervalApply,
    DateTimeIntervalApplyList,
    DateTimeIntervalList,
)
from ._date_transformation import (
    DateTransformation,
    DateTransformationApply,
    DateTransformationApplyList,
    DateTransformationList,
)
from ._day_ahead_bid import DayAheadBid, DayAheadBidApply, DayAheadBidApplyList, DayAheadBidList
from ._day_ahead_process import DayAheadProcess, DayAheadProcessApply, DayAheadProcessApplyList, DayAheadProcessList
from ._duration import Duration, DurationApply, DurationApplyList, DurationList
from ._generator import Generator, GeneratorApply, GeneratorApplyList, GeneratorList
from ._input_time_series_mapping import (
    InputTimeSeriesMapping,
    InputTimeSeriesMappingApply,
    InputTimeSeriesMappingApplyList,
    InputTimeSeriesMappingList,
)
from ._market_agreement import MarketAgreement, MarketAgreementApply, MarketAgreementApplyList, MarketAgreementList
from ._market_participant import (
    MarketParticipant,
    MarketParticipantApply,
    MarketParticipantApplyList,
    MarketParticipantList,
)
from ._mba_domain import MBADomain, MBADomainApply, MBADomainApplyList, MBADomainList
from ._nord_pool_market import NordPoolMarket, NordPoolMarketApply, NordPoolMarketApplyList, NordPoolMarketList
from ._output_container import OutputContainer, OutputContainerApply, OutputContainerApplyList, OutputContainerList
from ._output_mapping import OutputMapping, OutputMappingApply, OutputMappingApplyList, OutputMappingList
from ._plant import Plant, PlantApply, PlantApplyList, PlantList
from ._point import Point, PointApply, PointApplyList, PointList
from ._price_area import PriceArea, PriceAreaApply, PriceAreaApplyList, PriceAreaList
from ._production_plan_time_series import (
    ProductionPlanTimeSeries,
    ProductionPlanTimeSeriesApply,
    ProductionPlanTimeSeriesApplyList,
    ProductionPlanTimeSeriesList,
)
from ._reason import Reason, ReasonApply, ReasonApplyList, ReasonList
from ._reserve_bid import ReserveBid, ReserveBidApply, ReserveBidApplyList, ReserveBidList
from ._reserve_scenario import ReserveScenario, ReserveScenarioApply, ReserveScenarioApplyList, ReserveScenarioList
from ._reservoir import Reservoir, ReservoirApply, ReservoirApplyList, ReservoirList
from ._rkom_bid import RKOMBid, RKOMBidApply, RKOMBidApplyList, RKOMBidList
from ._rkom_bid_combination import (
    RKOMBidCombination,
    RKOMBidCombinationApply,
    RKOMBidCombinationApplyList,
    RKOMBidCombinationList,
)
from ._rkom_combination_bid import (
    RKOMCombinationBid,
    RKOMCombinationBidApply,
    RKOMCombinationBidApplyList,
    RKOMCombinationBidList,
)
from ._rkom_market import RKOMMarket, RKOMMarketApply, RKOMMarketApplyList, RKOMMarketList
from ._rkom_process import RKOMProcess, RKOMProcessApply, RKOMProcessApplyList, RKOMProcessList
from ._scenario import Scenario, ScenarioApply, ScenarioApplyList, ScenarioList
from ._scenario_mapping import ScenarioMapping, ScenarioMappingApply, ScenarioMappingApplyList, ScenarioMappingList
from ._scenario_template import ScenarioTemplate, ScenarioTemplateApply, ScenarioTemplateApplyList, ScenarioTemplateList
from ._series import Series, SeriesApply, SeriesApplyList, SeriesList
from ._shop_transformation import (
    ShopTransformation,
    ShopTransformationApply,
    ShopTransformationApplyList,
    ShopTransformationList,
)
from ._value_transformation import (
    ValueTransformation,
    ValueTransformationApply,
    ValueTransformationApplyList,
    ValueTransformationList,
)
from ._watercourse import Watercourse, WatercourseApply, WatercourseApplyList, WatercourseList
from ._watercourse_shop import WatercourseShop, WatercourseShopApply, WatercourseShopApplyList, WatercourseShopList

BenchmarkBidApply.model_rebuild()
BenchmarkProcessApply.model_rebuild()
BidTimeSeriesApply.model_rebuild()
DayAheadBidApply.model_rebuild()
DayAheadProcessApply.model_rebuild()
InputTimeSeriesMappingApply.model_rebuild()
OutputContainerApply.model_rebuild()
PlantApply.model_rebuild()
PriceAreaApply.model_rebuild()
RKOMBidApply.model_rebuild()
RKOMBidCombinationApply.model_rebuild()
RKOMProcessApply.model_rebuild()
ReserveBidApply.model_rebuild()
ScenarioApply.model_rebuild()
ScenarioMappingApply.model_rebuild()
ScenarioTemplateApply.model_rebuild()
SeriesApply.model_rebuild()
ShopTransformationApply.model_rebuild()
WatercourseApply.model_rebuild()

__all__ = [
    "BenchmarkBid",
    "BenchmarkBidApply",
    "BenchmarkBidList",
    "BenchmarkBidApplyList",
    "BenchmarkProcess",
    "BenchmarkProcessApply",
    "BenchmarkProcessList",
    "BenchmarkProcessApplyList",
    "BidMatrixGenerator",
    "BidMatrixGeneratorApply",
    "BidMatrixGeneratorList",
    "BidMatrixGeneratorApplyList",
    "BidTimeSeries",
    "BidTimeSeriesApply",
    "BidTimeSeriesList",
    "BidTimeSeriesApplyList",
    "CommandConfig",
    "CommandConfigApply",
    "CommandConfigList",
    "CommandConfigApplyList",
    "DateTimeInterval",
    "DateTimeIntervalApply",
    "DateTimeIntervalList",
    "DateTimeIntervalApplyList",
    "DateTransformation",
    "DateTransformationApply",
    "DateTransformationList",
    "DateTransformationApplyList",
    "DayAheadBid",
    "DayAheadBidApply",
    "DayAheadBidList",
    "DayAheadBidApplyList",
    "DayAheadProcess",
    "DayAheadProcessApply",
    "DayAheadProcessList",
    "DayAheadProcessApplyList",
    "Duration",
    "DurationApply",
    "DurationList",
    "DurationApplyList",
    "Generator",
    "GeneratorApply",
    "GeneratorList",
    "GeneratorApplyList",
    "InputTimeSeriesMapping",
    "InputTimeSeriesMappingApply",
    "InputTimeSeriesMappingList",
    "InputTimeSeriesMappingApplyList",
    "MBADomain",
    "MBADomainApply",
    "MBADomainList",
    "MBADomainApplyList",
    "MarketAgreement",
    "MarketAgreementApply",
    "MarketAgreementList",
    "MarketAgreementApplyList",
    "MarketParticipant",
    "MarketParticipantApply",
    "MarketParticipantList",
    "MarketParticipantApplyList",
    "NordPoolMarket",
    "NordPoolMarketApply",
    "NordPoolMarketList",
    "NordPoolMarketApplyList",
    "OutputContainer",
    "OutputContainerApply",
    "OutputContainerList",
    "OutputContainerApplyList",
    "OutputMapping",
    "OutputMappingApply",
    "OutputMappingList",
    "OutputMappingApplyList",
    "Plant",
    "PlantApply",
    "PlantList",
    "PlantApplyList",
    "Point",
    "PointApply",
    "PointList",
    "PointApplyList",
    "PriceArea",
    "PriceAreaApply",
    "PriceAreaList",
    "PriceAreaApplyList",
    "ProductionPlanTimeSeries",
    "ProductionPlanTimeSeriesApply",
    "ProductionPlanTimeSeriesList",
    "ProductionPlanTimeSeriesApplyList",
    "RKOMBid",
    "RKOMBidApply",
    "RKOMBidList",
    "RKOMBidApplyList",
    "RKOMBidCombination",
    "RKOMBidCombinationApply",
    "RKOMBidCombinationList",
    "RKOMBidCombinationApplyList",
    "RKOMCombinationBid",
    "RKOMCombinationBidApply",
    "RKOMCombinationBidList",
    "RKOMCombinationBidApplyList",
    "RKOMMarket",
    "RKOMMarketApply",
    "RKOMMarketList",
    "RKOMMarketApplyList",
    "RKOMProcess",
    "RKOMProcessApply",
    "RKOMProcessList",
    "RKOMProcessApplyList",
    "Reason",
    "ReasonApply",
    "ReasonList",
    "ReasonApplyList",
    "ReserveBid",
    "ReserveBidApply",
    "ReserveBidList",
    "ReserveBidApplyList",
    "ReserveScenario",
    "ReserveScenarioApply",
    "ReserveScenarioList",
    "ReserveScenarioApplyList",
    "Reservoir",
    "ReservoirApply",
    "ReservoirList",
    "ReservoirApplyList",
    "Scenario",
    "ScenarioApply",
    "ScenarioList",
    "ScenarioApplyList",
    "ScenarioMapping",
    "ScenarioMappingApply",
    "ScenarioMappingList",
    "ScenarioMappingApplyList",
    "ScenarioTemplate",
    "ScenarioTemplateApply",
    "ScenarioTemplateList",
    "ScenarioTemplateApplyList",
    "Series",
    "SeriesApply",
    "SeriesList",
    "SeriesApplyList",
    "ShopTransformation",
    "ShopTransformationApply",
    "ShopTransformationList",
    "ShopTransformationApplyList",
    "ValueTransformation",
    "ValueTransformationApply",
    "ValueTransformationList",
    "ValueTransformationApplyList",
    "Watercourse",
    "WatercourseApply",
    "WatercourseList",
    "WatercourseApplyList",
    "WatercourseShop",
    "WatercourseShopApply",
    "WatercourseShopList",
    "WatercourseShopApplyList",
]



from __future__ import annotations

from pathlib import Path

from cognite.client import ClientConfig, CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.credentials import OAuthClientCredentials

from ._api.benchmark_bid import BenchmarkBidAPI
from ._api.benchmark_process import BenchmarkProcessAPI
from ._api.bid_matrix_generator import BidMatrixGeneratorAPI
from ._api.bid_time_series import BidTimeSeriesAPI
from ._api.command_config import CommandConfigAPI
from ._api.date_time_interval import DateTimeIntervalAPI
from ._api.date_transformation import DateTransformationAPI
from ._api.day_ahead_bid import DayAheadBidAPI
from ._api.day_ahead_process import DayAheadProcessAPI
from ._api.duration import DurationAPI
from ._api.generator import GeneratorAPI
from ._api.input_time_series_mapping import InputTimeSeriesMappingAPI
from ._api.market_agreement import MarketAgreementAPI
from ._api.market_participant import MarketParticipantAPI
from ._api.mba_domain import MBADomainAPI
from ._api.nord_pool_market import NordPoolMarketAPI
from ._api.output_container import OutputContainerAPI
from ._api.output_mapping import OutputMappingAPI
from ._api.plant import PlantAPI
from ._api.point import PointAPI
from ._api.price_area import PriceAreaAPI
from ._api.production_plan_time_series import ProductionPlanTimeSeriesAPI
from ._api.reason import ReasonAPI
from ._api.reserve_bid import ReserveBidAPI
from ._api.reserve_scenario import ReserveScenarioAPI
from ._api.reservoir import ReservoirAPI
from ._api.rkom_bid import RKOMBidAPI
from ._api.rkom_bid_combination import RKOMBidCombinationAPI
from ._api.rkom_combination_bid import RKOMCombinationBidAPI
from ._api.rkom_market import RKOMMarketAPI
from ._api.rkom_process import RKOMProcessAPI
from ._api.scenario import ScenarioAPI
from ._api.scenario_mapping import ScenarioMappingAPI
from ._api.scenario_template import ScenarioTemplateAPI
from ._api.series import SeriesAPI
from ._api.shop_transformation import ShopTransformationAPI
from ._api.value_transformation import ValueTransformationAPI
from ._api.watercourse import WatercourseAPI
from ._api.watercourse_shop import WatercourseShopAPI


class AFRRAPIs:
    """
    AFRRAPIs

    Data Model:
        space: power-ops
        externalId: afrrMarket
        version: 1

    """

    def __init__(self, client: CogniteClient):
        self.bid_time_series = BidTimeSeriesAPI(client, dm.ViewId("power-ops", "BidTimeSeries", "2c1787140841fd"))
        self.date_time_interval = DateTimeIntervalAPI(
            client, dm.ViewId("power-ops", "DateTimeInterval", "18f6639083458b")
        )
        self.duration = DurationAPI(client, dm.ViewId("power-ops", "Duration", "7433a3f6ac2be0"))
        self.mba_domain = MBADomainAPI(client, dm.ViewId("power-ops", "MarketAgreement", "815d42dc6e008d"))
        self.market_agreement = MarketAgreementAPI(
            client, dm.ViewId("power-ops", "MarketParticipant", "8c47d7b03faeda")
        )
        self.market_participant = MarketParticipantAPI(client, dm.ViewId("power-ops", "MBADomain", "9ac70d436c3313"))
        self.point = PointAPI(client, dm.ViewId("power-ops", "Point", "791cb15b0ae9e1"))
        self.reason = ReasonAPI(client, dm.ViewId("power-ops", "Reason", "d064355f848186"))
        self.reserve_bid = ReserveBidAPI(client, dm.ViewId("power-ops", "ReserveBid", "dfb50d1e05d2e5"))
        self.series = SeriesAPI(client, dm.ViewId("power-ops", "Series", "59d189398e78be"))


class BenchmarkAPIs:
    """
    BenchmarkAPIs

    Data Model:
        space: power-ops
        externalId: benchmarkMarket
        version: 1

    """

    def __init__(self, client: CogniteClient):
        self.benchmark_bid = BenchmarkBidAPI(client, dm.ViewId("power-ops", "BenchmarkBid", "cd2ea6a54b92a6"))
        self.benchmark_process = BenchmarkProcessAPI(
            client, dm.ViewId("power-ops", "BenchmarkProcess", "3c3a0761a5f084")
        )
        self.date_transformation = DateTransformationAPI(
            client, dm.ViewId("power-ops", "DateTransformation", "a7c71305ba1288")
        )
        self.input_time_series_mapping = InputTimeSeriesMappingAPI(
            client, dm.ViewId("power-ops", "InputTimeSeriesMapping", "2426123a688e61")
        )
        self.nord_pool_market = NordPoolMarketAPI(client, dm.ViewId("power-ops", "NordPoolMarket", "88c86032b9ac9c"))
        self.production_plan_time_series = ProductionPlanTimeSeriesAPI(
            client, dm.ViewId("power-ops", "ProductionPlanTimeSeries", "ca7ffcb6f63d3f")
        )
        self.scenario_mapping = ScenarioMappingAPI(client, dm.ViewId("power-ops", "ScenarioMapping", "e65d4465699308"))
        self.shop_transformation = ShopTransformationAPI(
            client, dm.ViewId("power-ops", "ShopTransformation", "a74d706d1bda99")
        )
        self.value_transformation = ValueTransformationAPI(
            client, dm.ViewId("power-ops", "ValueTransformation", "acd34e005f1986")
        )


class CogShopAPIs:
    """
    CogShopAPIs

    Data Model:
        space: power-ops
        externalId: cogshop
        version: 1

    """

    def __init__(self, client: CogniteClient):
        self.command_config = CommandConfigAPI(client, dm.ViewId("power-ops", "CommandConfig", "128f1e0abfc7c5"))
        self.input_time_series_mapping = InputTimeSeriesMappingAPI(
            client, dm.ViewId("power-ops", "InputTimeSeriesMapping", "39f0e93d6b2bc1")
        )
        self.output_container = OutputContainerAPI(client, dm.ViewId("power-ops", "OutputContainer", "ad054c0f19ea87"))
        self.output_mapping = OutputMappingAPI(client, dm.ViewId("power-ops", "OutputMapping", "58e6e8f0dadecc"))
        self.scenario = ScenarioAPI(client, dm.ViewId("power-ops", "Scenario", "eb6cd945bd1400"))
        self.scenario_mapping = ScenarioMappingAPI(client, dm.ViewId("power-ops", "ScenarioMapping", "1cd399b3faffc4"))
        self.scenario_template = ScenarioTemplateAPI(
            client, dm.ViewId("power-ops", "ScenarioTemplate", "77579c65a8cdf9")
        )
        self.value_transformation = ValueTransformationAPI(
            client, dm.ViewId("power-ops", "ValueTransformation", "1b641fef1e0a83")
        )


class DayAheadAPIs:
    """
    DayAheadAPIs

    Data Model:
        space: power-ops
        externalId: dayaheadMarket
        version: 1

    """

    def __init__(self, client: CogniteClient):
        self.bid_matrix_generator = BidMatrixGeneratorAPI(
            client, dm.ViewId("power-ops", "BidMatrixGenerator", "98145498689f24")
        )
        self.date_transformation = DateTransformationAPI(
            client, dm.ViewId("power-ops", "DateTransformation", "78995d48b59c57")
        )
        self.day_ahead_bid = DayAheadBidAPI(client, dm.ViewId("power-ops", "DayAheadBid", "bd0768f04d3708"))
        self.day_ahead_process = DayAheadProcessAPI(client, dm.ViewId("power-ops", "DayAheadProcess", "dd1bf62feefc9a"))
        self.input_time_series_mapping = InputTimeSeriesMappingAPI(
            client, dm.ViewId("power-ops", "InputTimeSeriesMapping", "9532a47c052eca")
        )
        self.nord_pool_market = NordPoolMarketAPI(client, dm.ViewId("power-ops", "NordPoolMarket", "919be6b14f829d"))
        self.scenario_mapping = ScenarioMappingAPI(client, dm.ViewId("power-ops", "ScenarioMapping", "7d1c17ee79d79d"))
        self.shop_transformation = ShopTransformationAPI(
            client, dm.ViewId("power-ops", "ShopTransformation", "2dd4c9f8e072b6")
        )
        self.value_transformation = ValueTransformationAPI(
            client, dm.ViewId("power-ops", "ValueTransformation", "72f3a548b93c67")
        )


class ProductionAPIs:
    """
    ProductionAPIs

    Data Model:
        space: power-ops
        externalId: production
        version: 1

    """

    def __init__(self, client: CogniteClient):
        self.generator = GeneratorAPI(client, dm.ViewId("power-ops", "Generator", "9178931bbaac71"))
        self.plant = PlantAPI(client, dm.ViewId("power-ops", "Plant", "836dcb3f5da1df"))
        self.price_area = PriceAreaAPI(client, dm.ViewId("power-ops", "PriceArea", "6849ae787cd368"))
        self.reservoir = ReservoirAPI(client, dm.ViewId("power-ops", "Reservoir", "3c822b0c3d68f7"))
        self.watercourse = WatercourseAPI(client, dm.ViewId("power-ops", "Watercourse", "96f5170f35ef70"))
        self.watercourse_shop = WatercourseShopAPI(client, dm.ViewId("power-ops", "WatercourseShop", "4b5321b1fccd06"))


class RKOMMarketAPIs:
    """
    RKOMMarketAPIs

    Data Model:
        space: power-ops
        externalId: rkomMarket
        version: 1

    """

    def __init__(self, client: CogniteClient):
        self.date_transformation = DateTransformationAPI(
            client, dm.ViewId("power-ops", "DateTransformation", "13820c127c31c0")
        )
        self.input_time_series_mapping = InputTimeSeriesMappingAPI(
            client, dm.ViewId("power-ops", "InputTimeSeriesMapping", "7fce8483d52568")
        )
        self.rkom_bid = RKOMBidAPI(client, dm.ViewId("power-ops", "ReserveScenario", "e971c10bd1e893"))
        self.rkom_bid_combination = RKOMBidCombinationAPI(client, dm.ViewId("power-ops", "RKOMBid", "5c1a2ba06aa41f"))
        self.rkom_combination_bid = RKOMCombinationBidAPI(
            client, dm.ViewId("power-ops", "RKOMBidCombination", "b8c2faf6e35abe")
        )
        self.rkom_market = RKOMMarketAPI(client, dm.ViewId("power-ops", "RKOMCombinationBid", "a81dbcbbcd26de"))
        self.rkom_process = RKOMProcessAPI(client, dm.ViewId("power-ops", "RKOMMarket", "c362cd4abb3d4e"))
        self.reserve_scenario = ReserveScenarioAPI(client, dm.ViewId("power-ops", "RKOMProcess", "268dee7a04a5c3"))
        self.scenario_mapping = ScenarioMappingAPI(client, dm.ViewId("power-ops", "ScenarioMapping", "2b5b1f6fa4f53d"))
        self.shop_transformation = ShopTransformationAPI(
            client, dm.ViewId("power-ops", "ShopTransformation", "d0a6c80379e55b")
        )
        self.value_transformation = ValueTransformationAPI(
            client, dm.ViewId("power-ops", "ValueTransformation", "894c9c530a1c1d")
        )


class GeneratedPowerOpsClient:
    """
    GeneratedPowerOpsClient

    Generated with:
        pygen = 0.21.1
        cognite-sdk = 6.27.0
        pydantic = 2.4.1

    """

    def __init__(self, config_or_client: CogniteClient | ClientConfig):
        if isinstance(config_or_client, CogniteClient):
            client = config_or_client
        elif isinstance(config_or_client, ClientConfig):
            client = CogniteClient(config_or_client)
        else:
            raise ValueError(f"Expected CogniteClient or ClientConfig, got {type(config_or_client)}")
        self.afrr = AFRRAPIs(client)
        self.benchmark = BenchmarkAPIs(client)
        self.cog_shop = CogShopAPIs(client)
        self.day_ahead = DayAheadAPIs(client)
        self.production = ProductionAPIs(client)
        self.rkom_market = RKOMMarketAPIs(client)

    @classmethod
    def azure_project(
        cls, tenant_id: str, client_id: str, client_secret: str, cdf_cluster: str, project: str
    ) -> GeneratedPowerOpsClient:
        credentials = OAuthClientCredentials.default_for_azure_ad(tenant_id, client_id, client_secret, cdf_cluster)
        config = ClientConfig.default(project, cdf_cluster, credentials)

        return cls(config)

    @classmethod
    def from_toml(cls, file_path: Path | str, section: str | None = "cognite") -> GeneratedPowerOpsClient:
        import toml

        toml_content = toml.load(file_path)
        if section is not None:
            try:
                toml_content = toml_content[section]
            except KeyError as e:
                raise ValueError(f"Could not find section '{section}' in {file_path}") from e

        return cls.azure_project(**toml_content)



from __future__ import annotations

import datetime
from collections.abc import Sequence
from typing import overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm

from cognite.powerops.client._generated.data_classes import (
    DateTimeInterval,
    DateTimeIntervalApply,
    DateTimeIntervalApplyList,
    DateTimeIntervalList,
)

from ._core import DEFAULT_LIMIT_READ, TypeAPI


class DateTimeIntervalAPI(TypeAPI[DateTimeInterval, DateTimeIntervalApply, DateTimeIntervalList]):
    def __init__(self, client: CogniteClient, view_id: dm.ViewId):
        super().__init__(
            client=client,
            sources=view_id,
            class_type=DateTimeInterval,
            class_apply_type=DateTimeIntervalApply,
            class_list=DateTimeIntervalList,
        )
        self.view_id = view_id

    def apply(
        self, date_time_interval: DateTimeIntervalApply | Sequence[DateTimeIntervalApply], replace: bool = False
    ) -> dm.InstancesApplyResult:
        if isinstance(date_time_interval, DateTimeIntervalApply):
            instances = date_time_interval.to_instances_apply()
        else:
            instances = DateTimeIntervalApplyList(date_time_interval).to_instances_apply()
        return self._client.data_modeling.instances.apply(nodes=instances.nodes, edges=instances.edges, replace=replace)

    def delete(self, external_id: str | Sequence[str]) -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(DateTimeIntervalApply.space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(DateTimeIntervalApply.space, id) for id in external_id],
            )

    @overload
    def retrieve(self, external_id: str) -> DateTimeInterval:
        ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> DateTimeIntervalList:
        ...

    def retrieve(self, external_id: str | Sequence[str]) -> DateTimeInterval | DateTimeIntervalList:
        if isinstance(external_id, str):
            return self._retrieve((self.sources.space, external_id))
        else:
            return self._retrieve([(self.sources.space, ext_id) for ext_id in external_id])

    def list(
        self,
        min_start: datetime.datetime | None = None,
        max_start: datetime.datetime | None = None,
        min_end: datetime.datetime | None = None,
        max_end: datetime.datetime | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> DateTimeIntervalList:
        filter_ = _create_filter(
            self.view_id,
            min_start,
            max_start,
            min_end,
            max_end,
            external_id_prefix,
            filter,
        )

        return self._list(limit=limit, filter=filter_)


def _create_filter(
    view_id: dm.ViewId,
    min_start: datetime.datetime | None = None,
    max_start: datetime.datetime | None = None,
    min_end: datetime.datetime | None = None,
    max_end: datetime.datetime | None = None,
    external_id_prefix: str | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if min_start or max_start:
        filters.append(
            dm.filters.Range(
                view_id.as_property_ref("start"),
                gte=min_start.isoformat() if min_start else None,
                lte=max_start.isoformat() if max_start else None,
            )
        )
    if min_end or max_end:
        filters.append(
            dm.filters.Range(
                view_id.as_property_ref("end"),
                gte=min_end.isoformat() if min_end else None,
                lte=max_end.isoformat() if max_end else None,
            )
        )
    if external_id_prefix:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

from typing import ClassVar, Optional

from pydantic import Field
from typing_extensions import TypeAlias

from cognite.powerops.client.data_classes import (
    BenchmarkBidApply,
    BenchmarkProcessApply,
    BidMatrixGeneratorApply,
    DayAheadBidApply,
    DayAheadProcessApply,
    NordPoolMarketApply,
    RKOMBidApply,
    RKOMBidCombinationApply,
    RKOMMarketApply,
    RKOMProcessApply,
)
from cognite.powerops.client.powerops_client import PowerOpsClient
from cognite.powerops.resync.models.base import DataModel, PowerOpsGraphQLModel, T_Model

from .graphql_schemas import GRAPHQL_MODELS

ExternalID: TypeAlias = str


class BenchmarkMarketDataModel(DataModel):
    graph_ql: ClassVar[PowerOpsGraphQLModel] = GRAPHQL_MODELS["benchmark"]
    benchmarking: list[BenchmarkProcessApply] = Field(default_factory=list)
    bids: dict[ExternalID, BenchmarkBidApply] = Field(default_factory=dict)

    @classmethod
    def from_cdf(
        cls: type[T_Model], client: PowerOpsClient, fetch_metadata: bool = True, fetch_content: bool = False
    ) -> T_Model:
        raise NotImplementedError()

    def standardize(self) -> None:
        raise NotImplementedError()


class DayAheadMarketDataModel(DataModel):
    graph_ql: ClassVar[PowerOpsGraphQLModel] = GRAPHQL_MODELS["dayahead"]
    dayahead_processes: list[DayAheadProcessApply] = Field(default_factory=list)
    bids: dict[ExternalID, DayAheadBidApply] = Field(default_factory=dict)
    bid_matrix_generator: dict[ExternalID, BidMatrixGeneratorApply] = Field(default_factory=dict)
    nordpool_market: Optional[NordPoolMarketApply] = None

    @classmethod
    def from_cdf(
        cls: type[T_Model], client: PowerOpsClient, fetch_metadata: bool = True, fetch_content: bool = False
    ) -> T_Model:
        raise NotImplementedError()

    def standardize(self) -> None:
        raise NotImplementedError()


class RKOMMarketDataModel(DataModel):
    graph_ql: ClassVar[PowerOpsGraphQLModel] = GRAPHQL_MODELS["rkom"]
    rkom_market: Optional[RKOMMarketApply] = None
    bids: dict[ExternalID, RKOMBidApply] = Field(default_factory=dict)
    rkom_bid_combinations: list[RKOMBidCombinationApply] = Field(default_factory=list)
    rkom_processes: list[RKOMProcessApply] = Field(default_factory=list)

    @classmethod
    def from_cdf(
        cls: type[T_Model], client: PowerOpsClient, fetch_metadata: bool = True, fetch_content: bool = False
    ) -> T_Model:
        raise NotImplementedError()

    def standardize(self) -> None:
        raise NotImplementedError()


class AFRRMarket(DataModel):
    graph_ql: ClassVar[PowerOpsGraphQLModel] = GRAPHQL_MODELS["aFRR"]

    @classmethod
    def from_cdf(
        cls: type[T_Model], client: PowerOpsClient, fetch_metadata: bool = True, fetch_content: bool = False
    ) -> T_Model:
        raise NotImplementedError()

    def standardize(self) -> None:
        raise NotImplementedError()



from __future__ import annotations

from collections.abc import Sequence
from typing import overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm

from cognite.powerops.client._generated.cogshop1.data_classes import (
    Transformation,
    TransformationApply,
    TransformationApplyList,
    TransformationList,
)

from ._core import DEFAULT_LIMIT_READ, TypeAPI


class TransformationAPI(TypeAPI[Transformation, TransformationApply, TransformationList]):
    def __init__(self, client: CogniteClient, view_id: dm.ViewId):
        super().__init__(
            client=client,
            sources=view_id,
            class_type=Transformation,
            class_apply_type=TransformationApply,
            class_list=TransformationList,
        )
        self.view_id = view_id

    def apply(
        self, transformation: TransformationApply | Sequence[TransformationApply], replace: bool = False
    ) -> dm.InstancesApplyResult:
        if isinstance(transformation, TransformationApply):
            instances = transformation.to_instances_apply()
        else:
            instances = TransformationApplyList(transformation).to_instances_apply()
        return self._client.data_modeling.instances.apply(nodes=instances.nodes, edges=instances.edges, replace=replace)

    def delete(self, external_id: str | Sequence[str]) -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(TransformationApply.space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(TransformationApply.space, id) for id in external_id],
            )

    @overload
    def retrieve(self, external_id: str) -> Transformation:
        ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> TransformationList:
        ...

    def retrieve(self, external_id: str | Sequence[str]) -> Transformation | TransformationList:
        if isinstance(external_id, str):
            return self._retrieve((self.sources.space, external_id))
        else:
            return self._retrieve([(self.sources.space, ext_id) for ext_id in external_id])

    def list(
        self,
        method: str | list[str] | None = None,
        method_prefix: str | None = None,
        arguments: str | list[str] | None = None,
        arguments_prefix: str | None = None,
        min_order: int | None = None,
        max_order: int | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> TransformationList:
        filter_ = _create_filter(
            self.view_id,
            method,
            method_prefix,
            arguments,
            arguments_prefix,
            min_order,
            max_order,
            external_id_prefix,
            filter,
        )

        return self._list(limit=limit, filter=filter_)


def _create_filter(
    view_id: dm.ViewId,
    method: str | list[str] | None = None,
    method_prefix: str | None = None,
    arguments: str | list[str] | None = None,
    arguments_prefix: str | None = None,
    min_order: int | None = None,
    max_order: int | None = None,
    external_id_prefix: str | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if method and isinstance(method, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("method"), value=method))
    if method and isinstance(method, list):
        filters.append(dm.filters.In(view_id.as_property_ref("method"), values=method))
    if method_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("method"), value=method_prefix))
    if arguments and isinstance(arguments, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("arguments"), value=arguments))
    if arguments and isinstance(arguments, list):
        filters.append(dm.filters.In(view_id.as_property_ref("arguments"), values=arguments))
    if arguments_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("arguments"), value=arguments_prefix))
    if min_order or max_order:
        filters.append(dm.filters.Range(view_id.as_property_ref("order"), gte=min_order, lte=max_order))
    if external_id_prefix:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import DomainModel, DomainModelApply, TypeApplyList, TypeList

if TYPE_CHECKING:
    from ._benchmark_bid import BenchmarkBidApply
    from ._production_plan_time_series import ProductionPlanTimeSeriesApply
    from ._shop_transformation import ShopTransformationApply

__all__ = ["BenchmarkProcess", "BenchmarkProcessApply", "BenchmarkProcessList", "BenchmarkProcessApplyList"]


class BenchmarkProcess(DomainModel):
    space: ClassVar[str] = "power-ops"
    name: Optional[str] = None
    metrics: Optional[dict] = None
    bid: Optional[str] = None
    shop: Optional[str] = None
    run_events: Optional[list[str]] = Field(None, alias="runEvents")
    production_plan_time_series: Optional[list[str]] = Field(None, alias="productionPlanTimeSeries")

    def as_apply(self) -> BenchmarkProcessApply:
        return BenchmarkProcessApply(
            external_id=self.external_id,
            name=self.name,
            metrics=self.metrics,
            bid=self.bid,
            shop=self.shop,
            run_events=self.run_events,
            production_plan_time_series=self.production_plan_time_series,
        )


class BenchmarkProcessApply(DomainModelApply):
    space: ClassVar[str] = "power-ops"
    name: Optional[str] = None
    metrics: Optional[dict] = None
    bid: Union[BenchmarkBidApply, str, None] = Field(None, repr=False)
    shop: Union[ShopTransformationApply, str, None] = Field(None, repr=False)
    run_events: Optional[list[str]] = None
    production_plan_time_series: Union[list[ProductionPlanTimeSeriesApply], list[str], None] = Field(
        default=None, repr=False
    )

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        properties = {}
        if self.name is not None:
            properties["name"] = self.name
        if self.metrics is not None:
            properties["metrics"] = self.metrics
        if self.bid is not None:
            properties["bid"] = {
                "space": "power-ops",
                "externalId": self.bid if isinstance(self.bid, str) else self.bid.external_id,
            }
        if self.shop is not None:
            properties["shop"] = {
                "space": "power-ops",
                "externalId": self.shop if isinstance(self.shop, str) else self.shop.external_id,
            }
        if self.run_events is not None:
            properties["runEvents"] = self.run_events
        if properties:
            source = dm.NodeOrEdgeData(
                source=dm.ContainerId("power-ops", "BenchmarkProcess"),
                properties=properties,
            )
            sources.append(source)
        if sources:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.existing_version,
                sources=sources,
            )
            nodes = [this_node]
        else:
            nodes = []

        edges = []
        cache.add(self.external_id)

        for production_plan_time_series in self.production_plan_time_series or []:
            edge = self._create_production_plan_time_series_edge(production_plan_time_series)
            if edge.external_id not in cache:
                edges.append(edge)
                cache.add(edge.external_id)

            if isinstance(production_plan_time_series, DomainModelApply):
                instances = production_plan_time_series._to_instances_apply(cache)
                nodes.extend(instances.nodes)
                edges.extend(instances.edges)

        if isinstance(self.bid, DomainModelApply):
            instances = self.bid._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        if isinstance(self.shop, DomainModelApply):
            instances = self.shop._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))

    def _create_production_plan_time_series_edge(
        self, production_plan_time_series: Union[str, ProductionPlanTimeSeriesApply]
    ) -> dm.EdgeApply:
        if isinstance(production_plan_time_series, str):
            end_node_ext_id = production_plan_time_series
        elif isinstance(production_plan_time_series, DomainModelApply):
            end_node_ext_id = production_plan_time_series.external_id
        else:
            raise TypeError(f"Expected str or ProductionPlanTimeSeriesApply, got {type(production_plan_time_series)}")

        return dm.EdgeApply(
            space="power-ops",
            external_id=f"{self.external_id}:{end_node_ext_id}",
            type=dm.DirectRelationReference("power-ops", "BenchmarkProcess.productionPlanTimeSeries"),
            start_node=dm.DirectRelationReference(self.space, self.external_id),
            end_node=dm.DirectRelationReference("power-ops", end_node_ext_id),
        )


class BenchmarkProcessList(TypeList[BenchmarkProcess]):
    _NODE = BenchmarkProcess

    def as_apply(self) -> BenchmarkProcessApplyList:
        return BenchmarkProcessApplyList([node.as_apply() for node in self.data])


class BenchmarkProcessApplyList(TypeApplyList[BenchmarkProcessApply]):
    _NODE = BenchmarkProcessApply



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import DomainModel, DomainModelApply, TypeApplyList, TypeList

if TYPE_CHECKING:
    from ._plant import PlantApply
    from ._watercourse_shop import WatercourseShopApply

__all__ = ["Watercourse", "WatercourseApply", "WatercourseList", "WatercourseApplyList"]


class Watercourse(DomainModel):
    space: ClassVar[str] = "power-ops"
    name: Optional[str] = None
    shop: Optional[str] = None
    production_obligation: Optional[list[str]] = Field(None, alias="productionObligation")
    plants: Optional[list[str]] = None

    def as_apply(self) -> WatercourseApply:
        return WatercourseApply(
            external_id=self.external_id,
            name=self.name,
            shop=self.shop,
            production_obligation=self.production_obligation,
            plants=self.plants,
        )


class WatercourseApply(DomainModelApply):
    space: ClassVar[str] = "power-ops"
    name: Optional[str] = None
    shop: Union[WatercourseShopApply, str, None] = Field(None, repr=False)
    production_obligation: Optional[list[str]] = None
    plants: Union[list[PlantApply], list[str], None] = Field(default=None, repr=False)

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        properties = {}
        if self.name is not None:
            properties["name"] = self.name
        if self.shop is not None:
            properties["shop"] = {
                "space": "power-ops",
                "externalId": self.shop if isinstance(self.shop, str) else self.shop.external_id,
            }
        if self.production_obligation is not None:
            properties["productionObligation"] = self.production_obligation
        if properties:
            source = dm.NodeOrEdgeData(
                source=dm.ContainerId("power-ops", "Watercourse"),
                properties=properties,
            )
            sources.append(source)
        if sources:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.existing_version,
                sources=sources,
            )
            nodes = [this_node]
        else:
            nodes = []

        edges = []
        cache.add(self.external_id)

        for plant in self.plants or []:
            edge = self._create_plant_edge(plant)
            if edge.external_id not in cache:
                edges.append(edge)
                cache.add(edge.external_id)

            if isinstance(plant, DomainModelApply):
                instances = plant._to_instances_apply(cache)
                nodes.extend(instances.nodes)
                edges.extend(instances.edges)

        if isinstance(self.shop, DomainModelApply):
            instances = self.shop._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))

    def _create_plant_edge(self, plant: Union[str, PlantApply]) -> dm.EdgeApply:
        if isinstance(plant, str):
            end_node_ext_id = plant
        elif isinstance(plant, DomainModelApply):
            end_node_ext_id = plant.external_id
        else:
            raise TypeError(f"Expected str or PlantApply, got {type(plant)}")

        return dm.EdgeApply(
            space="power-ops",
            external_id=f"{self.external_id}:{end_node_ext_id}",
            type=dm.DirectRelationReference("power-ops", "Watercourse.plants"),
            start_node=dm.DirectRelationReference(self.space, self.external_id),
            end_node=dm.DirectRelationReference("power-ops", end_node_ext_id),
        )


class WatercourseList(TypeList[Watercourse]):
    _NODE = Watercourse

    def as_apply(self) -> WatercourseApplyList:
        return WatercourseApplyList([node.as_apply() for node in self.data])


class WatercourseApplyList(TypeApplyList[WatercourseApply]):
    _NODE = WatercourseApply



from __future__ import annotations

from collections import defaultdict
from collections.abc import Sequence
from typing import overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm

from cognite.powerops.client._generated.data_classes import RKOMBid, RKOMBidApply, RKOMBidApplyList, RKOMBidList

from ._core import DEFAULT_LIMIT_READ, TypeAPI


class RKOMBidDateAPI:
    def __init__(self, client: CogniteClient):
        self._client = client

    def retrieve(self, external_id: str | Sequence[str]) -> dm.EdgeList:
        f = dm.filters
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "power-ops", "externalId": "RKOMBid.date"},
        )
        if isinstance(external_id, str):
            is_rkom_bid = f.Equals(
                ["edge", "startNode"],
                {"space": "power-ops", "externalId": external_id},
            )
            return self._client.data_modeling.instances.list("edge", limit=-1, filter=f.And(is_edge_type, is_rkom_bid))

        else:
            is_rkom_bids = f.In(
                ["edge", "startNode"],
                [{"space": "power-ops", "externalId": ext_id} for ext_id in external_id],
            )
            return self._client.data_modeling.instances.list("edge", limit=-1, filter=f.And(is_edge_type, is_rkom_bids))

    def list(self, rkom_bid_id: str | list[str] | None = None, limit=DEFAULT_LIMIT_READ) -> dm.EdgeList:
        f = dm.filters
        filters = []
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "power-ops", "externalId": "RKOMBid.date"},
        )
        filters.append(is_edge_type)
        if rkom_bid_id:
            rkom_bid_ids = [rkom_bid_id] if isinstance(rkom_bid_id, str) else rkom_bid_id
            is_rkom_bids = f.In(
                ["edge", "startNode"],
                [{"space": "power-ops", "externalId": ext_id} for ext_id in rkom_bid_ids],
            )
            filters.append(is_rkom_bids)

        return self._client.data_modeling.instances.list("edge", limit=limit, filter=f.And(*filters))


class RKOMBidPriceScenariosAPI:
    def __init__(self, client: CogniteClient):
        self._client = client

    def retrieve(self, external_id: str | Sequence[str]) -> dm.EdgeList:
        f = dm.filters
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "power-ops", "externalId": "RKOMBid.priceScenarios"},
        )
        if isinstance(external_id, str):
            is_rkom_bid = f.Equals(
                ["edge", "startNode"],
                {"space": "power-ops", "externalId": external_id},
            )
            return self._client.data_modeling.instances.list("edge", limit=-1, filter=f.And(is_edge_type, is_rkom_bid))

        else:
            is_rkom_bids = f.In(
                ["edge", "startNode"],
                [{"space": "power-ops", "externalId": ext_id} for ext_id in external_id],
            )
            return self._client.data_modeling.instances.list("edge", limit=-1, filter=f.And(is_edge_type, is_rkom_bids))

    def list(self, rkom_bid_id: str | list[str] | None = None, limit=DEFAULT_LIMIT_READ) -> dm.EdgeList:
        f = dm.filters
        filters = []
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "power-ops", "externalId": "RKOMBid.priceScenarios"},
        )
        filters.append(is_edge_type)
        if rkom_bid_id:
            rkom_bid_ids = [rkom_bid_id] if isinstance(rkom_bid_id, str) else rkom_bid_id
            is_rkom_bids = f.In(
                ["edge", "startNode"],
                [{"space": "power-ops", "externalId": ext_id} for ext_id in rkom_bid_ids],
            )
            filters.append(is_rkom_bids)

        return self._client.data_modeling.instances.list("edge", limit=limit, filter=f.And(*filters))


class RKOMBidReserveScenariosAPI:
    def __init__(self, client: CogniteClient):
        self._client = client

    def retrieve(self, external_id: str | Sequence[str]) -> dm.EdgeList:
        f = dm.filters
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "power-ops", "externalId": "RKOMBid.reserveScenarios"},
        )
        if isinstance(external_id, str):
            is_rkom_bid = f.Equals(
                ["edge", "startNode"],
                {"space": "power-ops", "externalId": external_id},
            )
            return self._client.data_modeling.instances.list("edge", limit=-1, filter=f.And(is_edge_type, is_rkom_bid))

        else:
            is_rkom_bids = f.In(
                ["edge", "startNode"],
                [{"space": "power-ops", "externalId": ext_id} for ext_id in external_id],
            )
            return self._client.data_modeling.instances.list("edge", limit=-1, filter=f.And(is_edge_type, is_rkom_bids))

    def list(self, rkom_bid_id: str | list[str] | None = None, limit=DEFAULT_LIMIT_READ) -> dm.EdgeList:
        f = dm.filters
        filters = []
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "power-ops", "externalId": "RKOMBid.reserveScenarios"},
        )
        filters.append(is_edge_type)
        if rkom_bid_id:
            rkom_bid_ids = [rkom_bid_id] if isinstance(rkom_bid_id, str) else rkom_bid_id
            is_rkom_bids = f.In(
                ["edge", "startNode"],
                [{"space": "power-ops", "externalId": ext_id} for ext_id in rkom_bid_ids],
            )
            filters.append(is_rkom_bids)

        return self._client.data_modeling.instances.list("edge", limit=limit, filter=f.And(*filters))


class RKOMBidAPI(TypeAPI[RKOMBid, RKOMBidApply, RKOMBidList]):
    def __init__(self, client: CogniteClient, view_id: dm.ViewId):
        super().__init__(
            client=client,
            sources=view_id,
            class_type=RKOMBid,
            class_apply_type=RKOMBidApply,
            class_list=RKOMBidList,
        )
        self.view_id = view_id
        self.date = RKOMBidDateAPI(client)
        self.price_scenarios = RKOMBidPriceScenariosAPI(client)
        self.reserve_scenarios = RKOMBidReserveScenariosAPI(client)

    def apply(self, rkom_bid: RKOMBidApply | Sequence[RKOMBidApply], replace: bool = False) -> dm.InstancesApplyResult:
        if isinstance(rkom_bid, RKOMBidApply):
            instances = rkom_bid.to_instances_apply()
        else:
            instances = RKOMBidApplyList(rkom_bid).to_instances_apply()
        return self._client.data_modeling.instances.apply(nodes=instances.nodes, edges=instances.edges, replace=replace)

    def delete(self, external_id: str | Sequence[str]) -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(RKOMBidApply.space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(RKOMBidApply.space, id) for id in external_id],
            )

    @overload
    def retrieve(self, external_id: str) -> RKOMBid:
        ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> RKOMBidList:
        ...

    def retrieve(self, external_id: str | Sequence[str]) -> RKOMBid | RKOMBidList:
        if isinstance(external_id, str):
            rkom_bid = self._retrieve((self.sources.space, external_id))

            date_edges = self.date.retrieve(external_id)
            rkom_bid.date = [edge.end_node.external_id for edge in date_edges]
            price_scenario_edges = self.price_scenarios.retrieve(external_id)
            rkom_bid.price_scenarios = [edge.end_node.external_id for edge in price_scenario_edges]
            reserve_scenario_edges = self.reserve_scenarios.retrieve(external_id)
            rkom_bid.reserve_scenarios = [edge.end_node.external_id for edge in reserve_scenario_edges]

            return rkom_bid
        else:
            rkom_bids = self._retrieve([(self.sources.space, ext_id) for ext_id in external_id])

            date_edges = self.date.retrieve(external_id)
            self._set_date(rkom_bids, date_edges)
            price_scenario_edges = self.price_scenarios.retrieve(external_id)
            self._set_price_scenarios(rkom_bids, price_scenario_edges)
            reserve_scenario_edges = self.reserve_scenarios.retrieve(external_id)
            self._set_reserve_scenarios(rkom_bids, reserve_scenario_edges)

            return rkom_bids

    def list(
        self,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        method: str | list[str] | None = None,
        method_prefix: str | None = None,
        min_minimum_price: float | None = None,
        max_minimum_price: float | None = None,
        min_price_premium: float | None = None,
        max_price_premium: float | None = None,
        watercourse: str | list[str] | None = None,
        watercourse_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
        retrieve_edges: bool = True,
    ) -> RKOMBidList:
        filter_ = _create_filter(
            self.view_id,
            name,
            name_prefix,
            method,
            method_prefix,
            min_minimum_price,
            max_minimum_price,
            min_price_premium,
            max_price_premium,
            watercourse,
            watercourse_prefix,
            external_id_prefix,
            filter,
        )

        rkom_bids = self._list(limit=limit, filter=filter_)

        if retrieve_edges:
            date_edges = self.date.list(rkom_bids.as_external_ids(), limit=-1)
            self._set_date(rkom_bids, date_edges)
            price_scenario_edges = self.price_scenarios.list(rkom_bids.as_external_ids(), limit=-1)
            self._set_price_scenarios(rkom_bids, price_scenario_edges)
            reserve_scenario_edges = self.reserve_scenarios.list(rkom_bids.as_external_ids(), limit=-1)
            self._set_reserve_scenarios(rkom_bids, reserve_scenario_edges)

        return rkom_bids

    @staticmethod
    def _set_date(rkom_bids: Sequence[RKOMBid], date_edges: Sequence[dm.Edge]):
        edges_by_start_node: dict[tuple, list] = defaultdict(list)
        for edge in date_edges:
            edges_by_start_node[edge.start_node.as_tuple()].append(edge)

        for rkom_bid in rkom_bids:
            node_id = rkom_bid.id_tuple()
            if node_id in edges_by_start_node:
                rkom_bid.date = [edge.end_node.external_id for edge in edges_by_start_node[node_id]]

    @staticmethod
    def _set_price_scenarios(rkom_bids: Sequence[RKOMBid], price_scenario_edges: Sequence[dm.Edge]):
        edges_by_start_node: dict[tuple, list] = defaultdict(list)
        for edge in price_scenario_edges:
            edges_by_start_node[edge.start_node.as_tuple()].append(edge)

        for rkom_bid in rkom_bids:
            node_id = rkom_bid.id_tuple()
            if node_id in edges_by_start_node:
                rkom_bid.price_scenarios = [edge.end_node.external_id for edge in edges_by_start_node[node_id]]

    @staticmethod
    def _set_reserve_scenarios(rkom_bids: Sequence[RKOMBid], reserve_scenario_edges: Sequence[dm.Edge]):
        edges_by_start_node: dict[tuple, list] = defaultdict(list)
        for edge in reserve_scenario_edges:
            edges_by_start_node[edge.start_node.as_tuple()].append(edge)

        for rkom_bid in rkom_bids:
            node_id = rkom_bid.id_tuple()
            if node_id in edges_by_start_node:
                rkom_bid.reserve_scenarios = [edge.end_node.external_id for edge in edges_by_start_node[node_id]]


def _create_filter(
    view_id: dm.ViewId,
    name: str | list[str] | None = None,
    name_prefix: str | None = None,
    method: str | list[str] | None = None,
    method_prefix: str | None = None,
    min_minimum_price: float | None = None,
    max_minimum_price: float | None = None,
    min_price_premium: float | None = None,
    max_price_premium: float | None = None,
    watercourse: str | list[str] | None = None,
    watercourse_prefix: str | None = None,
    external_id_prefix: str | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if name and isinstance(name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("name"), value=name))
    if name and isinstance(name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("name"), values=name))
    if name_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("name"), value=name_prefix))
    if method and isinstance(method, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("method"), value=method))
    if method and isinstance(method, list):
        filters.append(dm.filters.In(view_id.as_property_ref("method"), values=method))
    if method_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("method"), value=method_prefix))
    if min_minimum_price or max_minimum_price:
        filters.append(
            dm.filters.Range(view_id.as_property_ref("minimumPrice"), gte=min_minimum_price, lte=max_minimum_price)
        )
    if min_price_premium or max_price_premium:
        filters.append(
            dm.filters.Range(view_id.as_property_ref("pricePremium"), gte=min_price_premium, lte=max_price_premium)
        )
    if watercourse and isinstance(watercourse, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("watercourse"), value=watercourse))
    if watercourse and isinstance(watercourse, list):
        filters.append(dm.filters.In(view_id.as_property_ref("watercourse"), values=watercourse))
    if watercourse_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("watercourse"), value=watercourse_prefix))
    if external_id_prefix:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from ._case import Case, CaseApply, CaseApplyList, CaseList
from ._commands_config import CommandsConfig, CommandsConfigApply, CommandsConfigApplyList, CommandsConfigList
from ._file_ref import FileRef, FileRefApply, FileRefApplyList, FileRefList
from ._mapping import Mapping, MappingApply, MappingApplyList, MappingList
from ._model_template import ModelTemplate, ModelTemplateApply, ModelTemplateApplyList, ModelTemplateList
from ._processing_log import ProcessingLog, ProcessingLogApply, ProcessingLogApplyList, ProcessingLogList
from ._scenario import Scenario, ScenarioApply, ScenarioApplyList, ScenarioList
from ._transformation import Transformation, TransformationApply, TransformationApplyList, TransformationList

CaseApply.model_rebuild()
MappingApply.model_rebuild()
ModelTemplateApply.model_rebuild()
ScenarioApply.model_rebuild()

__all__ = [
    "Case",
    "CaseApply",
    "CaseList",
    "CaseApplyList",
    "CommandsConfig",
    "CommandsConfigApply",
    "CommandsConfigList",
    "CommandsConfigApplyList",
    "FileRef",
    "FileRefApply",
    "FileRefList",
    "FileRefApplyList",
    "Mapping",
    "MappingApply",
    "MappingList",
    "MappingApplyList",
    "ModelTemplate",
    "ModelTemplateApply",
    "ModelTemplateList",
    "ModelTemplateApplyList",
    "ProcessingLog",
    "ProcessingLogApply",
    "ProcessingLogList",
    "ProcessingLogApplyList",
    "Scenario",
    "ScenarioApply",
    "ScenarioList",
    "ScenarioApplyList",
    "Transformation",
    "TransformationApply",
    "TransformationList",
    "TransformationApplyList",
]



from __future__ import annotations

from collections.abc import Sequence
from typing import overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm

from cognite.powerops.client._generated.data_classes import (
    ReserveScenario,
    ReserveScenarioApply,
    ReserveScenarioApplyList,
    ReserveScenarioList,
)

from ._core import DEFAULT_LIMIT_READ, TypeAPI


class ReserveScenarioAPI(TypeAPI[ReserveScenario, ReserveScenarioApply, ReserveScenarioList]):
    def __init__(self, client: CogniteClient, view_id: dm.ViewId):
        super().__init__(
            client=client,
            sources=view_id,
            class_type=ReserveScenario,
            class_apply_type=ReserveScenarioApply,
            class_list=ReserveScenarioList,
        )
        self.view_id = view_id

    def apply(
        self, reserve_scenario: ReserveScenarioApply | Sequence[ReserveScenarioApply], replace: bool = False
    ) -> dm.InstancesApplyResult:
        if isinstance(reserve_scenario, ReserveScenarioApply):
            instances = reserve_scenario.to_instances_apply()
        else:
            instances = ReserveScenarioApplyList(reserve_scenario).to_instances_apply()
        return self._client.data_modeling.instances.apply(nodes=instances.nodes, edges=instances.edges, replace=replace)

    def delete(self, external_id: str | Sequence[str]) -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(ReserveScenarioApply.space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(ReserveScenarioApply.space, id) for id in external_id],
            )

    @overload
    def retrieve(self, external_id: str) -> ReserveScenario:
        ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> ReserveScenarioList:
        ...

    def retrieve(self, external_id: str | Sequence[str]) -> ReserveScenario | ReserveScenarioList:
        if isinstance(external_id, str):
            return self._retrieve((self.sources.space, external_id))
        else:
            return self._retrieve([(self.sources.space, ext_id) for ext_id in external_id])

    def list(
        self,
        min_volume: int | None = None,
        max_volume: int | None = None,
        auction: str | list[str] | None = None,
        auction_prefix: str | None = None,
        product: str | list[str] | None = None,
        product_prefix: str | None = None,
        block: str | list[str] | None = None,
        block_prefix: str | None = None,
        reserve_group: str | list[str] | None = None,
        reserve_group_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> ReserveScenarioList:
        filter_ = _create_filter(
            self.view_id,
            min_volume,
            max_volume,
            auction,
            auction_prefix,
            product,
            product_prefix,
            block,
            block_prefix,
            reserve_group,
            reserve_group_prefix,
            external_id_prefix,
            filter,
        )

        return self._list(limit=limit, filter=filter_)


def _create_filter(
    view_id: dm.ViewId,
    min_volume: int | None = None,
    max_volume: int | None = None,
    auction: str | list[str] | None = None,
    auction_prefix: str | None = None,
    product: str | list[str] | None = None,
    product_prefix: str | None = None,
    block: str | list[str] | None = None,
    block_prefix: str | None = None,
    reserve_group: str | list[str] | None = None,
    reserve_group_prefix: str | None = None,
    external_id_prefix: str | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if min_volume or max_volume:
        filters.append(dm.filters.Range(view_id.as_property_ref("volume"), gte=min_volume, lte=max_volume))
    if auction and isinstance(auction, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("auction"), value=auction))
    if auction and isinstance(auction, list):
        filters.append(dm.filters.In(view_id.as_property_ref("auction"), values=auction))
    if auction_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("auction"), value=auction_prefix))
    if product and isinstance(product, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("product"), value=product))
    if product and isinstance(product, list):
        filters.append(dm.filters.In(view_id.as_property_ref("product"), values=product))
    if product_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("product"), value=product_prefix))
    if block and isinstance(block, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("block"), value=block))
    if block and isinstance(block, list):
        filters.append(dm.filters.In(view_id.as_property_ref("block"), values=block))
    if block_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("block"), value=block_prefix))
    if reserve_group and isinstance(reserve_group, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("reserveGroup"), value=reserve_group))
    if reserve_group and isinstance(reserve_group, list):
        filters.append(dm.filters.In(view_id.as_property_ref("reserveGroup"), values=reserve_group))
    if reserve_group_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("reserveGroup"), value=reserve_group_prefix))
    if external_id_prefix:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, ClassVar, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import DomainModel, DomainModelApply, TypeApplyList, TypeList

if TYPE_CHECKING:
    from ._market_participant import MarketParticipantApply
    from ._mba_domain import MBADomainApply

__all__ = ["ReserveBid", "ReserveBidApply", "ReserveBidList", "ReserveBidApplyList"]


class ReserveBid(DomainModel):
    space: ClassVar[str] = "power-ops"
    m_rid: Optional[str] = Field(None, alias="mRID")
    revision_number: Optional[str] = Field(None, alias="revisionNumber")
    type: Optional[str] = None
    process_type: Optional[str] = Field(None, alias="processType")
    sender: Optional[str] = None
    receiver: Optional[str] = None
    created_date_time: Optional[datetime.datetime] = Field(None, alias="createdDateTime")
    domain: Optional[str] = None
    subject: Optional[str] = None

    def as_apply(self) -> ReserveBidApply:
        return ReserveBidApply(
            external_id=self.external_id,
            m_rid=self.m_rid,
            revision_number=self.revision_number,
            type=self.type,
            process_type=self.process_type,
            sender=self.sender,
            receiver=self.receiver,
            created_date_time=self.created_date_time,
            domain=self.domain,
            subject=self.subject,
        )


class ReserveBidApply(DomainModelApply):
    space: ClassVar[str] = "power-ops"
    m_rid: Optional[str] = None
    revision_number: Optional[str] = None
    type: Optional[str] = None
    process_type: Optional[str] = None
    sender: Union[MarketParticipantApply, str, None] = Field(None, repr=False)
    receiver: Union[MarketParticipantApply, str, None] = Field(None, repr=False)
    created_date_time: Optional[datetime.datetime] = None
    domain: Union[MBADomainApply, str, None] = Field(None, repr=False)
    subject: Union[MarketParticipantApply, str, None] = Field(None, repr=False)

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        properties = {}
        if self.m_rid is not None:
            properties["mRID"] = self.m_rid
        if self.revision_number is not None:
            properties["revisionNumber"] = self.revision_number
        if self.type is not None:
            properties["type"] = self.type
        if self.process_type is not None:
            properties["processType"] = self.process_type
        if self.sender is not None:
            properties["sender"] = {
                "space": "power-ops",
                "externalId": self.sender if isinstance(self.sender, str) else self.sender.external_id,
            }
        if self.receiver is not None:
            properties["receiver"] = {
                "space": "power-ops",
                "externalId": self.receiver if isinstance(self.receiver, str) else self.receiver.external_id,
            }
        if self.created_date_time is not None:
            properties["createdDateTime"] = self.created_date_time.isoformat()
        if self.domain is not None:
            properties["domain"] = {
                "space": "power-ops",
                "externalId": self.domain if isinstance(self.domain, str) else self.domain.external_id,
            }
        if self.subject is not None:
            properties["subject"] = {
                "space": "power-ops",
                "externalId": self.subject if isinstance(self.subject, str) else self.subject.external_id,
            }
        if properties:
            source = dm.NodeOrEdgeData(
                source=dm.ContainerId("power-ops", "ReserveBid"),
                properties=properties,
            )
            sources.append(source)
        if sources:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.existing_version,
                sources=sources,
            )
            nodes = [this_node]
        else:
            nodes = []

        edges = []
        cache.add(self.external_id)

        if isinstance(self.sender, DomainModelApply):
            instances = self.sender._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        if isinstance(self.receiver, DomainModelApply):
            instances = self.receiver._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        if isinstance(self.domain, DomainModelApply):
            instances = self.domain._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        if isinstance(self.subject, DomainModelApply):
            instances = self.subject._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))


class ReserveBidList(TypeList[ReserveBid]):
    _NODE = ReserveBid

    def as_apply(self) -> ReserveBidApplyList:
        return ReserveBidApplyList([node.as_apply() for node in self.data])


class ReserveBidApplyList(TypeApplyList[ReserveBidApply]):
    _NODE = ReserveBidApply



from __future__ import annotations

from collections.abc import Sequence
from typing import overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm

from cognite.powerops.client._generated.data_classes import (
    NordPoolMarket,
    NordPoolMarketApply,
    NordPoolMarketApplyList,
    NordPoolMarketList,
)

from ._core import DEFAULT_LIMIT_READ, TypeAPI


class NordPoolMarketAPI(TypeAPI[NordPoolMarket, NordPoolMarketApply, NordPoolMarketList]):
    def __init__(self, client: CogniteClient, view_id: dm.ViewId):
        super().__init__(
            client=client,
            sources=view_id,
            class_type=NordPoolMarket,
            class_apply_type=NordPoolMarketApply,
            class_list=NordPoolMarketList,
        )
        self.view_id = view_id

    def apply(
        self, nord_pool_market: NordPoolMarketApply | Sequence[NordPoolMarketApply], replace: bool = False
    ) -> dm.InstancesApplyResult:
        if isinstance(nord_pool_market, NordPoolMarketApply):
            instances = nord_pool_market.to_instances_apply()
        else:
            instances = NordPoolMarketApplyList(nord_pool_market).to_instances_apply()
        return self._client.data_modeling.instances.apply(nodes=instances.nodes, edges=instances.edges, replace=replace)

    def delete(self, external_id: str | Sequence[str]) -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(NordPoolMarketApply.space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(NordPoolMarketApply.space, id) for id in external_id],
            )

    @overload
    def retrieve(self, external_id: str) -> NordPoolMarket:
        ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> NordPoolMarketList:
        ...

    def retrieve(self, external_id: str | Sequence[str]) -> NordPoolMarket | NordPoolMarketList:
        if isinstance(external_id, str):
            return self._retrieve((self.sources.space, external_id))
        else:
            return self._retrieve([(self.sources.space, ext_id) for ext_id in external_id])

    def list(
        self,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        timezone: str | list[str] | None = None,
        timezone_prefix: str | None = None,
        min_max_price: float | None = None,
        max_max_price: float | None = None,
        min_min_price: float | None = None,
        max_min_price: float | None = None,
        min_price_steps: int | None = None,
        max_price_steps: int | None = None,
        price_unit: str | list[str] | None = None,
        price_unit_prefix: str | None = None,
        min_tick_size: float | None = None,
        max_tick_size: float | None = None,
        time_unit: str | list[str] | None = None,
        time_unit_prefix: str | None = None,
        min_trade_lot: float | None = None,
        max_trade_lot: float | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> NordPoolMarketList:
        filter_ = _create_filter(
            self.view_id,
            name,
            name_prefix,
            timezone,
            timezone_prefix,
            min_max_price,
            max_max_price,
            min_min_price,
            max_min_price,
            min_price_steps,
            max_price_steps,
            price_unit,
            price_unit_prefix,
            min_tick_size,
            max_tick_size,
            time_unit,
            time_unit_prefix,
            min_trade_lot,
            max_trade_lot,
            external_id_prefix,
            filter,
        )

        return self._list(limit=limit, filter=filter_)


def _create_filter(
    view_id: dm.ViewId,
    name: str | list[str] | None = None,
    name_prefix: str | None = None,
    timezone: str | list[str] | None = None,
    timezone_prefix: str | None = None,
    min_max_price: float | None = None,
    max_max_price: float | None = None,
    min_min_price: float | None = None,
    max_min_price: float | None = None,
    min_price_steps: int | None = None,
    max_price_steps: int | None = None,
    price_unit: str | list[str] | None = None,
    price_unit_prefix: str | None = None,
    min_tick_size: float | None = None,
    max_tick_size: float | None = None,
    time_unit: str | list[str] | None = None,
    time_unit_prefix: str | None = None,
    min_trade_lot: float | None = None,
    max_trade_lot: float | None = None,
    external_id_prefix: str | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if name and isinstance(name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("name"), value=name))
    if name and isinstance(name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("name"), values=name))
    if name_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("name"), value=name_prefix))
    if timezone and isinstance(timezone, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("timezone"), value=timezone))
    if timezone and isinstance(timezone, list):
        filters.append(dm.filters.In(view_id.as_property_ref("timezone"), values=timezone))
    if timezone_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("timezone"), value=timezone_prefix))
    if min_max_price or max_max_price:
        filters.append(dm.filters.Range(view_id.as_property_ref("maxPrice"), gte=min_max_price, lte=max_max_price))
    if min_min_price or max_min_price:
        filters.append(dm.filters.Range(view_id.as_property_ref("minPrice"), gte=min_min_price, lte=max_min_price))
    if min_price_steps or max_price_steps:
        filters.append(
            dm.filters.Range(view_id.as_property_ref("priceSteps"), gte=min_price_steps, lte=max_price_steps)
        )
    if price_unit and isinstance(price_unit, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("priceUnit"), value=price_unit))
    if price_unit and isinstance(price_unit, list):
        filters.append(dm.filters.In(view_id.as_property_ref("priceUnit"), values=price_unit))
    if price_unit_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("priceUnit"), value=price_unit_prefix))
    if min_tick_size or max_tick_size:
        filters.append(dm.filters.Range(view_id.as_property_ref("tickSize"), gte=min_tick_size, lte=max_tick_size))
    if time_unit and isinstance(time_unit, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("timeUnit"), value=time_unit))
    if time_unit and isinstance(time_unit, list):
        filters.append(dm.filters.In(view_id.as_property_ref("timeUnit"), values=time_unit))
    if time_unit_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("timeUnit"), value=time_unit_prefix))
    if min_trade_lot or max_trade_lot:
        filters.append(dm.filters.Range(view_id.as_property_ref("tradeLot"), gte=min_trade_lot, lte=max_trade_lot))
    if external_id_prefix:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

from collections.abc import Sequence
from typing import overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm

from cognite.powerops.client._generated.data_classes import (
    BidTimeSeries,
    BidTimeSeriesApply,
    BidTimeSeriesApplyList,
    BidTimeSeriesList,
)

from ._core import DEFAULT_LIMIT_READ, TypeAPI


class BidTimeSeriesAPI(TypeAPI[BidTimeSeries, BidTimeSeriesApply, BidTimeSeriesList]):
    def __init__(self, client: CogniteClient, view_id: dm.ViewId):
        super().__init__(
            client=client,
            sources=view_id,
            class_type=BidTimeSeries,
            class_apply_type=BidTimeSeriesApply,
            class_list=BidTimeSeriesList,
        )
        self.view_id = view_id

    def apply(
        self, bid_time_series: BidTimeSeriesApply | Sequence[BidTimeSeriesApply], replace: bool = False
    ) -> dm.InstancesApplyResult:
        if isinstance(bid_time_series, BidTimeSeriesApply):
            instances = bid_time_series.to_instances_apply()
        else:
            instances = BidTimeSeriesApplyList(bid_time_series).to_instances_apply()
        return self._client.data_modeling.instances.apply(nodes=instances.nodes, edges=instances.edges, replace=replace)

    def delete(self, external_id: str | Sequence[str]) -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(BidTimeSeriesApply.space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(BidTimeSeriesApply.space, id) for id in external_id],
            )

    @overload
    def retrieve(self, external_id: str) -> BidTimeSeries:
        ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> BidTimeSeriesList:
        ...

    def retrieve(self, external_id: str | Sequence[str]) -> BidTimeSeries | BidTimeSeriesList:
        if isinstance(external_id, str):
            return self._retrieve((self.sources.space, external_id))
        else:
            return self._retrieve([(self.sources.space, ext_id) for ext_id in external_id])

    def list(
        self,
        m_rid: str | list[str] | None = None,
        m_rid_prefix: str | None = None,
        auction: str | list[str] | None = None,
        auction_prefix: str | None = None,
        quantity_measure_unit_name: str | list[str] | None = None,
        quantity_measure_unit_name_prefix: str | None = None,
        currency_unit_name: str | list[str] | None = None,
        currency_unit_name_prefix: str | None = None,
        price_measure_unit_name: str | list[str] | None = None,
        price_measure_unit_name_prefix: str | None = None,
        divisible: bool | None = None,
        block_bid: bool | None = None,
        min_status: int | None = None,
        max_status: int | None = None,
        min_priority: int | None = None,
        max_priority: int | None = None,
        registered_resources_mrid: str | list[str] | None = None,
        registered_resources_mrid_prefix: str | None = None,
        flow_direction: str | list[str] | None = None,
        flow_direction_prefix: str | None = None,
        min_step_increment_quantity: float | None = None,
        max_step_increment_quantity: float | None = None,
        energy_price_measure_unit: str | list[str] | None = None,
        energy_price_measure_unit_prefix: str | None = None,
        standard_market_product_type: str | list[str] | None = None,
        standard_market_product_type_prefix: str | None = None,
        original_market_product_type: str | list[str] | None = None,
        original_market_product_type_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> BidTimeSeriesList:
        filter_ = _create_filter(
            self.view_id,
            m_rid,
            m_rid_prefix,
            auction,
            auction_prefix,
            quantity_measure_unit_name,
            quantity_measure_unit_name_prefix,
            currency_unit_name,
            currency_unit_name_prefix,
            price_measure_unit_name,
            price_measure_unit_name_prefix,
            divisible,
            block_bid,
            min_status,
            max_status,
            min_priority,
            max_priority,
            registered_resources_mrid,
            registered_resources_mrid_prefix,
            flow_direction,
            flow_direction_prefix,
            min_step_increment_quantity,
            max_step_increment_quantity,
            energy_price_measure_unit,
            energy_price_measure_unit_prefix,
            standard_market_product_type,
            standard_market_product_type_prefix,
            original_market_product_type,
            original_market_product_type_prefix,
            external_id_prefix,
            filter,
        )

        return self._list(limit=limit, filter=filter_)


def _create_filter(
    view_id: dm.ViewId,
    m_rid: str | list[str] | None = None,
    m_rid_prefix: str | None = None,
    auction: str | list[str] | None = None,
    auction_prefix: str | None = None,
    quantity_measure_unit_name: str | list[str] | None = None,
    quantity_measure_unit_name_prefix: str | None = None,
    currency_unit_name: str | list[str] | None = None,
    currency_unit_name_prefix: str | None = None,
    price_measure_unit_name: str | list[str] | None = None,
    price_measure_unit_name_prefix: str | None = None,
    divisible: bool | None = None,
    block_bid: bool | None = None,
    min_status: int | None = None,
    max_status: int | None = None,
    min_priority: int | None = None,
    max_priority: int | None = None,
    registered_resources_mrid: str | list[str] | None = None,
    registered_resources_mrid_prefix: str | None = None,
    flow_direction: str | list[str] | None = None,
    flow_direction_prefix: str | None = None,
    min_step_increment_quantity: float | None = None,
    max_step_increment_quantity: float | None = None,
    energy_price_measure_unit: str | list[str] | None = None,
    energy_price_measure_unit_prefix: str | None = None,
    standard_market_product_type: str | list[str] | None = None,
    standard_market_product_type_prefix: str | None = None,
    original_market_product_type: str | list[str] | None = None,
    original_market_product_type_prefix: str | None = None,
    external_id_prefix: str | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if m_rid and isinstance(m_rid, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("mRID"), value=m_rid))
    if m_rid and isinstance(m_rid, list):
        filters.append(dm.filters.In(view_id.as_property_ref("mRID"), values=m_rid))
    if m_rid_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("mRID"), value=m_rid_prefix))
    if auction and isinstance(auction, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("auction"), value=auction))
    if auction and isinstance(auction, list):
        filters.append(dm.filters.In(view_id.as_property_ref("auction"), values=auction))
    if auction_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("auction"), value=auction_prefix))
    if quantity_measure_unit_name and isinstance(quantity_measure_unit_name, str):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("quantityMeasureUnitName"), value=quantity_measure_unit_name)
        )
    if quantity_measure_unit_name and isinstance(quantity_measure_unit_name, list):
        filters.append(
            dm.filters.In(view_id.as_property_ref("quantityMeasureUnitName"), values=quantity_measure_unit_name)
        )
    if quantity_measure_unit_name_prefix:
        filters.append(
            dm.filters.Prefix(
                view_id.as_property_ref("quantityMeasureUnitName"), value=quantity_measure_unit_name_prefix
            )
        )
    if currency_unit_name and isinstance(currency_unit_name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("currencyUnitName"), value=currency_unit_name))
    if currency_unit_name and isinstance(currency_unit_name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("currencyUnitName"), values=currency_unit_name))
    if currency_unit_name_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("currencyUnitName"), value=currency_unit_name_prefix))
    if price_measure_unit_name and isinstance(price_measure_unit_name, str):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("priceMeasureUnitName"), value=price_measure_unit_name)
        )
    if price_measure_unit_name and isinstance(price_measure_unit_name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("priceMeasureUnitName"), values=price_measure_unit_name))
    if price_measure_unit_name_prefix:
        filters.append(
            dm.filters.Prefix(view_id.as_property_ref("priceMeasureUnitName"), value=price_measure_unit_name_prefix)
        )
    if divisible and isinstance(divisible, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("divisible"), value=divisible))
    if block_bid and isinstance(block_bid, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("blockBid"), value=block_bid))
    if min_status or max_status:
        filters.append(dm.filters.Range(view_id.as_property_ref("status"), gte=min_status, lte=max_status))
    if min_priority or max_priority:
        filters.append(dm.filters.Range(view_id.as_property_ref("priority"), gte=min_priority, lte=max_priority))
    if registered_resources_mrid and isinstance(registered_resources_mrid, str):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("registeredResourcesMRID"), value=registered_resources_mrid)
        )
    if registered_resources_mrid and isinstance(registered_resources_mrid, list):
        filters.append(
            dm.filters.In(view_id.as_property_ref("registeredResourcesMRID"), values=registered_resources_mrid)
        )
    if registered_resources_mrid_prefix:
        filters.append(
            dm.filters.Prefix(
                view_id.as_property_ref("registeredResourcesMRID"), value=registered_resources_mrid_prefix
            )
        )
    if flow_direction and isinstance(flow_direction, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("flowDirection"), value=flow_direction))
    if flow_direction and isinstance(flow_direction, list):
        filters.append(dm.filters.In(view_id.as_property_ref("flowDirection"), values=flow_direction))
    if flow_direction_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("flowDirection"), value=flow_direction_prefix))
    if min_step_increment_quantity or max_step_increment_quantity:
        filters.append(
            dm.filters.Range(
                view_id.as_property_ref("stepIncrementQuantity"),
                gte=min_step_increment_quantity,
                lte=max_step_increment_quantity,
            )
        )
    if energy_price_measure_unit and isinstance(energy_price_measure_unit, str):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("energyPriceMeasureUnit"), value=energy_price_measure_unit)
        )
    if energy_price_measure_unit and isinstance(energy_price_measure_unit, list):
        filters.append(
            dm.filters.In(view_id.as_property_ref("energyPriceMeasureUnit"), values=energy_price_measure_unit)
        )
    if energy_price_measure_unit_prefix:
        filters.append(
            dm.filters.Prefix(view_id.as_property_ref("energyPriceMeasureUnit"), value=energy_price_measure_unit_prefix)
        )
    if standard_market_product_type and isinstance(standard_market_product_type, str):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("standardMarketProductType"), value=standard_market_product_type)
        )
    if standard_market_product_type and isinstance(standard_market_product_type, list):
        filters.append(
            dm.filters.In(view_id.as_property_ref("standardMarketProductType"), values=standard_market_product_type)
        )
    if standard_market_product_type_prefix:
        filters.append(
            dm.filters.Prefix(
                view_id.as_property_ref("standardMarketProductType"), value=standard_market_product_type_prefix
            )
        )
    if original_market_product_type and isinstance(original_market_product_type, str):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("originalMarketProductType"), value=original_market_product_type)
        )
    if original_market_product_type and isinstance(original_market_product_type, list):
        filters.append(
            dm.filters.In(view_id.as_property_ref("originalMarketProductType"), values=original_market_product_type)
        )
    if original_market_product_type_prefix:
        filters.append(
            dm.filters.Prefix(
                view_id.as_property_ref("originalMarketProductType"), value=original_market_product_type_prefix
            )
        )
    if external_id_prefix:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

import datetime
from collections import defaultdict
from collections.abc import Sequence
from typing import overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm

from cognite.powerops.client._generated.data_classes import Series, SeriesApply, SeriesApplyList, SeriesList

from ._core import DEFAULT_LIMIT_READ, TypeAPI


class SeriesPointsAPI:
    def __init__(self, client: CogniteClient):
        self._client = client

    def retrieve(self, external_id: str | Sequence[str]) -> dm.EdgeList:
        f = dm.filters
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "power-ops", "externalId": "Series.points"},
        )
        if isinstance(external_id, str):
            is_series = f.Equals(
                ["edge", "startNode"],
                {"space": "power-ops", "externalId": external_id},
            )
            return self._client.data_modeling.instances.list("edge", limit=-1, filter=f.And(is_edge_type, is_series))

        else:
            is_series_list = f.In(
                ["edge", "startNode"],
                [{"space": "power-ops", "externalId": ext_id} for ext_id in external_id],
            )
            return self._client.data_modeling.instances.list(
                "edge", limit=-1, filter=f.And(is_edge_type, is_series_list)
            )

    def list(self, series_id: str | list[str] | None = None, limit=DEFAULT_LIMIT_READ) -> dm.EdgeList:
        f = dm.filters
        filters = []
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "power-ops", "externalId": "Series.points"},
        )
        filters.append(is_edge_type)
        if series_id:
            series_ids = [series_id] if isinstance(series_id, str) else series_id
            is_series_list = f.In(
                ["edge", "startNode"],
                [{"space": "power-ops", "externalId": ext_id} for ext_id in series_ids],
            )
            filters.append(is_series_list)

        return self._client.data_modeling.instances.list("edge", limit=limit, filter=f.And(*filters))


class SeriesAPI(TypeAPI[Series, SeriesApply, SeriesList]):
    def __init__(self, client: CogniteClient, view_id: dm.ViewId):
        super().__init__(
            client=client,
            sources=view_id,
            class_type=Series,
            class_apply_type=SeriesApply,
            class_list=SeriesList,
        )
        self.view_id = view_id
        self.points = SeriesPointsAPI(client)

    def apply(self, series: SeriesApply | Sequence[SeriesApply], replace: bool = False) -> dm.InstancesApplyResult:
        if isinstance(series, SeriesApply):
            instances = series.to_instances_apply()
        else:
            instances = SeriesApplyList(series).to_instances_apply()
        return self._client.data_modeling.instances.apply(nodes=instances.nodes, edges=instances.edges, replace=replace)

    def delete(self, external_id: str | Sequence[str]) -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(SeriesApply.space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(SeriesApply.space, id) for id in external_id],
            )

    @overload
    def retrieve(self, external_id: str) -> Series:
        ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> SeriesList:
        ...

    def retrieve(self, external_id: str | Sequence[str]) -> Series | SeriesList:
        if isinstance(external_id, str):
            series = self._retrieve((self.sources.space, external_id))

            point_edges = self.points.retrieve(external_id)
            series.points = [edge.end_node.external_id for edge in point_edges]

            return series
        else:
            series_list = self._retrieve([(self.sources.space, ext_id) for ext_id in external_id])

            point_edges = self.points.retrieve(external_id)
            self._set_points(series_list, point_edges)

            return series_list

    def list(
        self,
        min_time_interval_start: datetime.datetime | None = None,
        max_time_interval_start: datetime.datetime | None = None,
        min_time_interval_end: datetime.datetime | None = None,
        max_time_interval_end: datetime.datetime | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
        retrieve_edges: bool = True,
    ) -> SeriesList:
        filter_ = _create_filter(
            self.view_id,
            min_time_interval_start,
            max_time_interval_start,
            min_time_interval_end,
            max_time_interval_end,
            external_id_prefix,
            filter,
        )

        series_list = self._list(limit=limit, filter=filter_)

        if retrieve_edges:
            point_edges = self.points.list(series_list.as_external_ids(), limit=-1)
            self._set_points(series_list, point_edges)

        return series_list

    @staticmethod
    def _set_points(series_list: Sequence[Series], point_edges: Sequence[dm.Edge]):
        edges_by_start_node: dict[tuple, list] = defaultdict(list)
        for edge in point_edges:
            edges_by_start_node[edge.start_node.as_tuple()].append(edge)

        for series in series_list:
            node_id = series.id_tuple()
            if node_id in edges_by_start_node:
                series.points = [edge.end_node.external_id for edge in edges_by_start_node[node_id]]


def _create_filter(
    view_id: dm.ViewId,
    min_time_interval_start: datetime.datetime | None = None,
    max_time_interval_start: datetime.datetime | None = None,
    min_time_interval_end: datetime.datetime | None = None,
    max_time_interval_end: datetime.datetime | None = None,
    external_id_prefix: str | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if min_time_interval_start or max_time_interval_start:
        filters.append(
            dm.filters.Range(
                view_id.as_property_ref("timeIntervalStart"),
                gte=min_time_interval_start.isoformat() if min_time_interval_start else None,
                lte=max_time_interval_start.isoformat() if max_time_interval_start else None,
            )
        )
    if min_time_interval_end or max_time_interval_end:
        filters.append(
            dm.filters.Range(
                view_id.as_property_ref("timeIntervalEnd"),
                gte=min_time_interval_end.isoformat() if min_time_interval_end else None,
                lte=max_time_interval_end.isoformat() if max_time_interval_end else None,
            )
        )
    if external_id_prefix:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

from collections import defaultdict
from collections.abc import Sequence
from typing import overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm

from cognite.powerops.client._generated.data_classes import (
    ShopTransformation,
    ShopTransformationApply,
    ShopTransformationApplyList,
    ShopTransformationList,
)

from ._core import DEFAULT_LIMIT_READ, TypeAPI


class ShopTransformationEndAPI:
    def __init__(self, client: CogniteClient):
        self._client = client

    def retrieve(self, external_id: str | Sequence[str]) -> dm.EdgeList:
        f = dm.filters
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "power-ops", "externalId": "ShopTransformation.end"},
        )
        if isinstance(external_id, str):
            is_shop_transformation = f.Equals(
                ["edge", "startNode"],
                {"space": "power-ops", "externalId": external_id},
            )
            return self._client.data_modeling.instances.list(
                "edge", limit=-1, filter=f.And(is_edge_type, is_shop_transformation)
            )

        else:
            is_shop_transformations = f.In(
                ["edge", "startNode"],
                [{"space": "power-ops", "externalId": ext_id} for ext_id in external_id],
            )
            return self._client.data_modeling.instances.list(
                "edge", limit=-1, filter=f.And(is_edge_type, is_shop_transformations)
            )

    def list(self, shop_transformation_id: str | list[str] | None = None, limit=DEFAULT_LIMIT_READ) -> dm.EdgeList:
        f = dm.filters
        filters = []
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "power-ops", "externalId": "ShopTransformation.end"},
        )
        filters.append(is_edge_type)
        if shop_transformation_id:
            shop_transformation_ids = (
                [shop_transformation_id] if isinstance(shop_transformation_id, str) else shop_transformation_id
            )
            is_shop_transformations = f.In(
                ["edge", "startNode"],
                [{"space": "power-ops", "externalId": ext_id} for ext_id in shop_transformation_ids],
            )
            filters.append(is_shop_transformations)

        return self._client.data_modeling.instances.list("edge", limit=limit, filter=f.And(*filters))


class ShopTransformationStartAPI:
    def __init__(self, client: CogniteClient):
        self._client = client

    def retrieve(self, external_id: str | Sequence[str]) -> dm.EdgeList:
        f = dm.filters
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "power-ops", "externalId": "ShopTransformation.start"},
        )
        if isinstance(external_id, str):
            is_shop_transformation = f.Equals(
                ["edge", "startNode"],
                {"space": "power-ops", "externalId": external_id},
            )
            return self._client.data_modeling.instances.list(
                "edge", limit=-1, filter=f.And(is_edge_type, is_shop_transformation)
            )

        else:
            is_shop_transformations = f.In(
                ["edge", "startNode"],
                [{"space": "power-ops", "externalId": ext_id} for ext_id in external_id],
            )
            return self._client.data_modeling.instances.list(
                "edge", limit=-1, filter=f.And(is_edge_type, is_shop_transformations)
            )

    def list(self, shop_transformation_id: str | list[str] | None = None, limit=DEFAULT_LIMIT_READ) -> dm.EdgeList:
        f = dm.filters
        filters = []
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "power-ops", "externalId": "ShopTransformation.start"},
        )
        filters.append(is_edge_type)
        if shop_transformation_id:
            shop_transformation_ids = (
                [shop_transformation_id] if isinstance(shop_transformation_id, str) else shop_transformation_id
            )
            is_shop_transformations = f.In(
                ["edge", "startNode"],
                [{"space": "power-ops", "externalId": ext_id} for ext_id in shop_transformation_ids],
            )
            filters.append(is_shop_transformations)

        return self._client.data_modeling.instances.list("edge", limit=limit, filter=f.And(*filters))


class ShopTransformationAPI(TypeAPI[ShopTransformation, ShopTransformationApply, ShopTransformationList]):
    def __init__(self, client: CogniteClient, view_id: dm.ViewId):
        super().__init__(
            client=client,
            sources=view_id,
            class_type=ShopTransformation,
            class_apply_type=ShopTransformationApply,
            class_list=ShopTransformationList,
        )
        self.view_id = view_id
        self.end = ShopTransformationEndAPI(client)
        self.start = ShopTransformationStartAPI(client)

    def apply(
        self, shop_transformation: ShopTransformationApply | Sequence[ShopTransformationApply], replace: bool = False
    ) -> dm.InstancesApplyResult:
        if isinstance(shop_transformation, ShopTransformationApply):
            instances = shop_transformation.to_instances_apply()
        else:
            instances = ShopTransformationApplyList(shop_transformation).to_instances_apply()
        return self._client.data_modeling.instances.apply(nodes=instances.nodes, edges=instances.edges, replace=replace)

    def delete(self, external_id: str | Sequence[str]) -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(ShopTransformationApply.space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(ShopTransformationApply.space, id) for id in external_id],
            )

    @overload
    def retrieve(self, external_id: str) -> ShopTransformation:
        ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> ShopTransformationList:
        ...

    def retrieve(self, external_id: str | Sequence[str]) -> ShopTransformation | ShopTransformationList:
        if isinstance(external_id, str):
            shop_transformation = self._retrieve((self.sources.space, external_id))

            end_edges = self.end.retrieve(external_id)
            shop_transformation.end = [edge.end_node.external_id for edge in end_edges]
            start_edges = self.start.retrieve(external_id)
            shop_transformation.start = [edge.end_node.external_id for edge in start_edges]

            return shop_transformation
        else:
            shop_transformations = self._retrieve([(self.sources.space, ext_id) for ext_id in external_id])

            end_edges = self.end.retrieve(external_id)
            self._set_end(shop_transformations, end_edges)
            start_edges = self.start.retrieve(external_id)
            self._set_start(shop_transformations, start_edges)

            return shop_transformations

    def list(
        self,
        type_name: str | list[str] | None = None,
        type_name_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
        retrieve_edges: bool = True,
    ) -> ShopTransformationList:
        filter_ = _create_filter(
            self.view_id,
            type_name,
            type_name_prefix,
            external_id_prefix,
            filter,
        )

        shop_transformations = self._list(limit=limit, filter=filter_)

        if retrieve_edges:
            end_edges = self.end.list(shop_transformations.as_external_ids(), limit=-1)
            self._set_end(shop_transformations, end_edges)
            start_edges = self.start.list(shop_transformations.as_external_ids(), limit=-1)
            self._set_start(shop_transformations, start_edges)

        return shop_transformations

    @staticmethod
    def _set_end(shop_transformations: Sequence[ShopTransformation], end_edges: Sequence[dm.Edge]):
        edges_by_start_node: dict[tuple, list] = defaultdict(list)
        for edge in end_edges:
            edges_by_start_node[edge.start_node.as_tuple()].append(edge)

        for shop_transformation in shop_transformations:
            node_id = shop_transformation.id_tuple()
            if node_id in edges_by_start_node:
                shop_transformation.end = [edge.end_node.external_id for edge in edges_by_start_node[node_id]]

    @staticmethod
    def _set_start(shop_transformations: Sequence[ShopTransformation], start_edges: Sequence[dm.Edge]):
        edges_by_start_node: dict[tuple, list] = defaultdict(list)
        for edge in start_edges:
            edges_by_start_node[edge.start_node.as_tuple()].append(edge)

        for shop_transformation in shop_transformations:
            node_id = shop_transformation.id_tuple()
            if node_id in edges_by_start_node:
                shop_transformation.start = [edge.end_node.external_id for edge in edges_by_start_node[node_id]]


def _create_filter(
    view_id: dm.ViewId,
    type_name: str | list[str] | None = None,
    type_name_prefix: str | None = None,
    external_id_prefix: str | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if type_name and isinstance(type_name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("typeName"), value=type_name))
    if type_name and isinstance(type_name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("typeName"), values=type_name))
    if type_name_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("typeName"), value=type_name_prefix))
    if external_id_prefix:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

from typing import ClassVar, Optional

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import DomainModel, DomainModelApply, TypeApplyList, TypeList

__all__ = ["BidMatrixGenerator", "BidMatrixGeneratorApply", "BidMatrixGeneratorList", "BidMatrixGeneratorApplyList"]


class BidMatrixGenerator(DomainModel):
    space: ClassVar[str] = "power-ops"
    shop_plant: Optional[str] = Field(None, alias="shopPlant")
    methods: Optional[str] = None
    function_external_id: Optional[str] = Field(None, alias="functionExternalId")

    def as_apply(self) -> BidMatrixGeneratorApply:
        return BidMatrixGeneratorApply(
            external_id=self.external_id,
            shop_plant=self.shop_plant,
            methods=self.methods,
            function_external_id=self.function_external_id,
        )


class BidMatrixGeneratorApply(DomainModelApply):
    space: ClassVar[str] = "power-ops"
    shop_plant: Optional[str] = None
    methods: Optional[str] = None
    function_external_id: Optional[str] = None

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        properties = {}
        if self.shop_plant is not None:
            properties["shopPlant"] = self.shop_plant
        if self.methods is not None:
            properties["methods"] = self.methods
        if self.function_external_id is not None:
            properties["functionExternalId"] = self.function_external_id
        if properties:
            source = dm.NodeOrEdgeData(
                source=dm.ContainerId("power-ops", "BidMatrixGenerator"),
                properties=properties,
            )
            sources.append(source)
        if sources:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.existing_version,
                sources=sources,
            )
            nodes = [this_node]
        else:
            nodes = []

        edges = []
        cache.add(self.external_id)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))


class BidMatrixGeneratorList(TypeList[BidMatrixGenerator]):
    _NODE = BidMatrixGenerator

    def as_apply(self) -> BidMatrixGeneratorApplyList:
        return BidMatrixGeneratorApplyList([node.as_apply() for node in self.data])


class BidMatrixGeneratorApplyList(TypeApplyList[BidMatrixGeneratorApply]):
    _NODE = BidMatrixGeneratorApply
